/***************************************************************************
                          client.c  -  description
                             -------------------
    begin                : Sun Apr 28 12:02:57 CEST 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <time.h>
#include "lbreakout.h"
#include "config.h"
#include "event.h"
#include "shrapnells.h"
#include "levels.h"
#include "game.h"
#include "value.h"
#include "item.h"
#include "menu.h"
#include "windows.h"
#include "client.h"
#include "gui.h"
#include "../common/parser.h"

/*
====================================================================
Externals
====================================================================
*/
extern Config config;
extern GUI *gui;
extern SDL_Surface *cr_wallpaper;
extern Sdl sdl;
extern Font *chat_font_error;
extern List *levels;
extern int levelset_version, levelset_update;

/*
====================================================================
Client
====================================================================
*/
typedef struct {
    int  id;
    char name[16];
} CChannel;
typedef struct {
    int  id;
    char name[16];
} CUser;
typedef struct {
    int id;
    char players[40]; /* challenger vs challenged */
    char set[16]; /* levelset name */
    int  host_wins, guest_wins; /* won rounds */
} CGame;
List *users = 0; /* users known to client */
List *games = 0; /* running games */
List *channels = 0; /* list of strings naming the channels */
char client_error[128]; /* error message */
#ifdef WITH_NETWORK
Net_Socket *client = 0; /* client socket to the game server */
#endif
int client_id; /* id assigned by server */
char client_name[16]; /* our local username */
char client_ip[NET_IP_SIZE]; /* local ip send when logging in to server */
SDL_Surface *cr_bkgnd = 0;
enum {
    CSTATE_NONE = 0,
    CSTATE_INFO,
    CSTATE_AWAIT_ANSWER, /* wait for answer to a challenge */
    CSTATE_ANSWER, /* answer to a challenge */
    CSTATE_CONFIRM_TRANSFER, /* say yes or no to transfer */
    CSTATE_AWAIT_TRANSFER_CONFIRMATION, /* wait for answer on transfer offer */
    CSTATE_RECEIVE /* receive level data */
};
int client_state = CSTATE_NONE;
/* CHALLENGE */
char *mp_diff_names[] = { "Easy", "Medium", "Hard" };
int mp_peer_id;
char mp_peer_name[16]; /* remote player we want to play with */
char mp_levelset[16]; /* levelset we play */
char mp_levelset_version[16]; /* version of levelset */
int mp_level_count; /* number of levels in set */
int mp_diff, mp_rounds, mp_frags, mp_balls; /* game configuration */

/*
====================================================================
Locals
====================================================================
*/

/*
====================================================================
Run some tests to benchmark the speed of your machine.
Return the time in milliseconds the benchmark took.
====================================================================
*/
int client_benchmark()
{
    SDL_Surface *surf = create_surf( 640, 480, SDL_SWSURFACE );
    int i, start;
    double var = 15364.32498745;
    double var2 = 278834.3121398243;
    start = SDL_GetTicks();
    /* some floating point operations */
    for ( i = 0; i < 100000; i++ )
        var = log(  exp( var ) );
    for ( i = 0; i < 100000; i++ ) {
        var *= var2;
        var /= var2;
    }
    /* blitting stuff */
    for ( i = 0; i < 10; i++ ) {
        FULL_DEST( surf ); FULL_SOURCE( sdl.screen ); blit_surf();
        FULL_DEST( sdl.screen ); FULL_SOURCE( surf ); alpha_blit_surf( 128 );
        FULL_DEST( sdl.screen ); FULL_SOURCE( surf ); blit_surf();
    }
    /* stop! */
    SDL_FreeSurface( surf );
    return SDL_GetTicks() - start;
}

/*
====================================================================
Add/remove a channel from list.
====================================================================
*/
void client_add_channel( int id, char *name )
{
    CChannel *channel = calloc( 1, sizeof( CChannel ) );
    if ( channel ) {
        channel->id = id;
        strcpy_lt( channel->name, name, 15 );
        list_add( channels, channel );
        list_add( gui->channels->entries, strdup( name ) );
    }
}
void client_remove_channel( int id )
{
    CChannel *channel;
    list_reset( channels );
    list_reset( gui->channels->entries );
    while ( ( channel = list_next( channels ) ) ) {
        list_next( gui->channels->entries );
        if ( channel->id == id ) {
            list_delete_current( channels );
            list_delete_current( gui->channels->entries );
            gui->channels->cur_item = 0;
        }
    }
}

/*
====================================================================
Find channel by name.
====================================================================
*/
CChannel *client_find_channel_by_name( char *name )
{
    CChannel *channel;
    list_reset( channels );
    while ( ( channel = list_next( channels ) ) )
        if ( STRCMP( channel->name, name ) )
            return channel;
    return 0;
}

/*
====================================================================
Return pointer to channel name in gui::channels by passed id
====================================================================
*/
char *client_find_channel_name( int id )
{
    char *name;
    CChannel *channel;
    list_reset( channels );
    list_reset( gui->channels->entries );
    while ( ( channel = list_next( channels ) ) && ( name = list_next( gui->channels->entries ) ) ) {
        if ( channel->id == id )
            return name;
    }
    return 0;
}

/*
====================================================================
Add/remove a user from the user list.
====================================================================
*/
CUser* client_find_user( int id )
{
    CUser *entry;
    list_reset( users );
    while ( ( entry = list_next( users ) ) )
        if ( entry->id == id )
            return entry;
    return 0;
}
int client_get_user_id( char *name )
{
    CUser *entry;
    list_reset( users );
    while ( ( entry = list_next( users ) ) )
        if ( STRCMP( entry->name, name ) )
            return entry->id;
    return 0;
}
void client_add_user( int id, char *name )
{
    CUser *user;
    if ( !client_find_user( id ) ) {
        user = calloc( 1, sizeof( CUser ) );
        if ( user ) {
            strcpy_lt( user->name, name, 12 );
            user->id = id;
            list_add( users, user );
            list_add( gui->users->entries, strdup( name ) );
        }
    }
}
void client_remove_user( int id ) {
    char *uname;
    CUser *entry = client_find_user( id );
    if ( entry ) {
        list_reset( gui->users->entries );
        while ( ( uname = list_next( gui->users->entries ) ) )
            if ( STRCMP( uname, entry->name ) ) {
                if ( gui->users->cur_item == uname )
                    gui->users->cur_item = 0;
                list_delete_item( gui->users->entries, uname );
                break;
            }
        list_delete_item( users, entry );
    }
}

/*
====================================================================
Add/remove/update a game in the game list.
====================================================================
*/
CGame *client_find_game( int id )
{
    CGame *entry = 0;
    list_reset( games );
    while ( ( entry = list_next( games ) ) )
        if ( entry->id == id )
            return entry;
    return 0;
}
void client_add_game( int id, char *players, char *set )
{
    CGame *game;
    char buf[48];
    if ( !client_find_game( id ) ) {
        game = calloc( 1, sizeof( CGame ) );
        if ( game ) {
            game->id = id;
            strcpy_lt( game->players, players, 39 );
            strcpy_lt( game->set, set, 15 );
            list_add( games, game );
            sprintf( buf, "%s%2i:%i", players, 0, 0 );
            list_add( gui->games->entries, strdup( buf ) );
        }
    }
}
void client_rebuilt_game_list()
{
    char buf[48];
    CGame *game = 0;
    list_clear( gui->games->entries );
    list_reset( games );
    while ( ( game = list_next( games ) ) ) {
        sprintf( buf, "%s%2i:%i", 
                 game->players, game->host_wins, game->guest_wins );
        list_add( gui->games->entries, strdup( buf ) );
    }
}
void client_remove_game( int id )
{
    CGame *game = client_find_game( id );
    if ( game ) {
        list_delete_item( games, game );
        client_rebuilt_game_list();
    }
}
void client_update_game( int id, int host_win )
{
    CGame *game = client_find_game( id );
    if ( game ) {
        game->host_wins += host_win;
        game->guest_wins += !host_win;
        client_rebuilt_game_list();
    }
}

/*
====================================================================
Clear channels, users and games for incoming update.
====================================================================
*/
void client_prepare_update()
{
    gui->channels->cur_item = 0;
    list_clear( channels );
    list_clear( gui->channels->entries );
    mp_peer_id = 0; mp_peer_name[0] = 0;
    gui->users->cur_item = 0;
    list_clear( users );
    list_clear( gui->users->entries );
    list_clear( games );
    list_clear( gui->games->entries );
}

/*
====================================================================
Open/close confirm window.
====================================================================
*/
void client_draw_confirm_text( char *str )
{
    int i, y = 10;
    Text *text = 0;
    SDL_Surface *contents = gui->confirm->frame->contents;
    SDL_FillRect( contents, 0, 0x0 );
            gui->font->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
    text = create_text( str, 25 );
    for ( i = 0; i < text->count; i++ )
        write_line( contents, gui->font, text->lines[i], 10, &y );
    delete_text( text );
    iframe_apply( gui->confirm->frame );
}
int client_popup( int type )
{
    SDL_Surface *contents = gui->confirm->frame->contents;
    char text[128];
    int y = 10;
    switch ( type ) {
        case CSTATE_CONFIRM_TRANSFER:
            group_set_active( gui->confirm, ID_OK, 1 );
            group_set_active( gui->confirm, ID_CANCEL, 1 );
            sprintf( text, "%s wants to send you the levelset %s v%s (%i levels). Do you accept?", 
                     mp_peer_name, mp_levelset, mp_levelset_version, mp_level_count );
            client_draw_confirm_text( text );
            group_hide( gui->confirm, 0 );
            client_state = type;
            break;
        case CSTATE_AWAIT_TRANSFER_CONFIRMATION:
            group_set_active( gui->confirm, ID_OK, 0 );
            group_set_active( gui->confirm, ID_CANCEL, 1 );
            sprintf( text, "You've offered the levelset %s to %s.#Awaiting confirmation...", 
                     mp_levelset, mp_peer_name );
            client_draw_confirm_text( text );
            group_hide( gui->confirm, 0 );
            client_state = type;
            break;
        case CSTATE_AWAIT_ANSWER:
            group_set_active( gui->confirm, ID_OK, 0 );
            group_set_active( gui->confirm, ID_CANCEL, 1 );
            sprintf( text, "You've challenged %s. Let's see what (s)he says...", 
                     mp_peer_name );
            client_draw_confirm_text( text );
            group_hide( gui->confirm, 0 );
            client_state = type;
            break;
        case CSTATE_ANSWER:
            SDL_FillRect( contents, 0, 0x0 );
            group_set_active( gui->confirm, ID_OK, 1 );
            group_set_active( gui->confirm, ID_CANCEL, 1 );
            gui->font->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
            write_line( contents, gui->font, "You have been challenged!", 10, &y ); y+= 10;
            sprintf( text, "Challenger: %13s", mp_peer_name );
            write_line( contents, gui->font, text, 10, &y );
            sprintf( text, "Levelset:   %13s", mp_levelset );
            write_line( contents, gui->font, text, 10, &y );
            sprintf( text, "Difficulty: %13s", mp_diff_names[mp_diff] );
            write_line( contents, gui->font, text, 10, &y );
            sprintf( text, "Rounds:     %13i", mp_rounds );
            write_line( contents, gui->font, text, 10, &y );
            sprintf( text, "Frag Limit: %13i", mp_frags );
            write_line( contents, gui->font, text, 10, &y );
            sprintf( text, "Balls:      %13i", mp_balls );
            write_line( contents, gui->font, text, 10, &y );
            iframe_apply( gui->confirm->frame );
            group_hide( gui->confirm, 0 );
            client_state = type;
            break;
    }
    return 1;
}
void client_popup_info( char *info )
{
    client_draw_confirm_text( info );
    group_set_active( gui->confirm, ID_OK, 0 );
    group_set_active( gui->confirm, ID_CANCEL, 1 );
    group_hide( gui->confirm, 0 );
    client_state = CSTATE_INFO;
}
void client_close_popup()
{
    group_hide( gui->confirm, 1 );
}

/*
====================================================================
Check if you want to start a transfer, load levels and popup info.
====================================================================
*/
int client_check_transfer( char *chatter )
{
#ifdef WITH_NETWORK
    Net_Msg msg;
    List *args;
    char *set, *name;
    char version[16];
    char buf[64];
    if ( client_state != CSTATE_NONE )
        return 0;
    if ( strlen( chatter ) >= 9 && !strncmp( "/transfer", chatter, 9 ) ) {
        args = parser_explode_string( chatter, ' ' ); list_next( args );
        if ( ( set = list_next( args ) ) == 0 )
            gui_add_chat_msg( gui->chat_win, CHAT_WIN_LINE_WIDTH, "!And what shall I send???", 0 );
        else
            if ( ( name = list_next( args ) ) == 0 )
                gui_add_chat_msg( gui->chat_win, CHAT_WIN_LINE_WIDTH, "!And to whom shall I send it???", 0 );
            else {
                if ( !levels_load( set ) ) {
                    sprintf( buf, "!Can't find levelset '%s'.", set );
                    gui_add_chat_msg( gui->chat_win, CHAT_WIN_LINE_WIDTH, buf, 0 );
                }
                else {
                    sprintf( version, "%i.%02i", levelset_version, levelset_update );
                    net_init_msg( &msg, MSG_INIT_TRANSFER );
                    if ( !net_pack_string( &msg, name ) ||
                         !net_pack_string( &msg, (set[0]=='~')?set+1:set ) ||
                         !net_pack_string( &msg, version ) ||
                         !net_pack_int16( &msg, levels->count ) ||
                         !net_write_msg( client, &msg ) )
                        fprintf( stderr, "%s\n", net_get_error() );
                    else {
                        strcpy_lt( mp_peer_name, name, 15 );
                        strcpy_lt( mp_levelset, set, 15 );
                        client_popup( CSTATE_AWAIT_TRANSFER_CONFIRMATION );
                    }
                }
                /* clear chat edit */
                gui->chat_edit->text[0] = 0;
                gui->chat_edit->length = 0;
                gui_edit_draw( gui->chat_edit );
            }
        list_delete( args );
        return 1;
    }
    return 0;
#else
    return 1;
#endif
}

/*
====================================================================
Establish a direct connection and run the game.
Return False if there was an error.
====================================================================
*/
void client_run_game( int challenger )
{
#ifdef WITH_NETWORK    
    char buf[128], ip_name[NET_IP_SIZE];
    int port, ip_recv = 0;
    Net_Msg msg;
    Net_Socket *gserver = 0, *endpoint = 0;
    int start_time;
    int host = 0; /* the host is the faster machine that
                     also builds the frames for the client */
    int my_result, remote_result; /* benchmark results */
    int bench_recv = 0;
    int game_port = atoi( config.client_game_port );
    fade_screen( FADE_OUT, FADE_DEF_TIME );
    /* clear error */
    client_error[0] = 0;
    /* info */
    if ( challenger )
        sprintf( buf, "Waiting for remote player..." );
    else
        sprintf( buf, "Connecting to remote player..." );
    gui->font->align = ALIGN_X_CENTER | ALIGN_Y_CENTER;
    write_text( gui->font, sdl.screen, sdl.screen->w/2, sdl.screen->h/2, buf, OPAQUE );
    refresh_screen( 0, 0, 0, 0 );
    /* make the connection */
    if ( challenger ) {
        /* the challenger opens a game server */
        for ( port = game_port; port < game_port + 10; port++ )
            if ( ( gserver = net_open( port, 0 ) ) )
                break;
        if ( gserver == 0 )
            sprintf( client_error, "%s", net_get_error() );
        else {
            /* send port and wait for answer */
            net_init_msg( &msg, MSG_IP_ADDRESS );
            if ( !net_pack_string( &msg, client_ip ) ||
                 !net_pack_int32( &msg, port ) ||
                 !net_write_msg( client, &msg ) )
                sprintf( client_error, "%s", net_get_error() );
            else {
                start_time = time( 0 );
                while ( time( 0 ) <= start_time + 8 )
                    if ( ( endpoint = net_accept( gserver ) ) )
                        break;
                if ( endpoint == 0 )
                    sprintf( client_error, "Remote player did not respond" );
            }
        }
    }
    else {
        /* the client receives the port and connects to it */
        if ( net_read_msg( client, &msg, 8000 ) )
            if ( msg.type == MSG_IP_ADDRESS ) {
                net_unpack_string( &msg, ip_name, NET_IP_SIZE );
                net_unpack_int32( &msg, &port );
#ifdef NET_INFO
                printf( "IP received: %s\n", ip_name );
                //strcpy( ip_name, "80.6.55.46" );
                printf( "IP forced: %s:%i\n", ip_name, port );
#endif
                ip_recv = 1;
            }
        if ( ip_recv ) {
            if ( ( endpoint = net_connect( ip_name, port ) ) == 0 )
                sprintf( client_error, net_get_error() );
        }
        else
            if ( client_error[0] == 0 )
                sprintf( client_error, "Challenger didn't send ip address (timeout)" );
    }
    /* check who's faster */
    if ( client_error[0] == 0 ) {
        my_result = client_benchmark();
        net_init_msg( &msg, MSG_BENCHMARK );
        net_pack_int16( &msg, my_result );
        net_write_msg( endpoint, &msg );
        if ( net_read_msg( endpoint, &msg, 8000 ) )
            if ( msg.type == MSG_BENCHMARK )
                if ( net_unpack_int16( &msg, &remote_result ) )
                    bench_recv = 1;
        if ( bench_recv ) {
            /* the one with less time wins: equal time decides 
               for the challenger */
            if ( my_result == remote_result )
                host = challenger;
            else
                if ( my_result < remote_result )
                    host = 1;
                else
                    host = 0;
                host = challenger;
            printf( "Local: %i, Remote: %i, Host: %i\n", 
                    my_result, remote_result, host );
        }
        else
            if ( client_error[0] == 0 )
                sprintf( client_error, "Remote player didn't send benchmark result (timeout)" );
    }
    /* run game if no error */
    if ( client_error[0] == 0 ) {
        if ( game_init( mp_levelset, mp_diff, endpoint, challenger, host, mp_rounds, mp_frags, mp_balls ) ) 
            game_run();
        game_clear();
        net_close( endpoint );
        if ( challenger )
            net_close( gserver );
    }
    if ( client_error[0] != 0 ) {
        /* show error */
        SDL_FillRect( sdl.screen, 0, 0x0 );
        chat_font_error->align = ALIGN_X_CENTER | ALIGN_Y_CENTER;
        write_text( chat_font_error, sdl.screen, sdl.screen->w/2, sdl.screen->h/2, client_error, OPAQUE );
        refresh_screen( 0, 0, 0, 0 );
        event_clear_sdl_queue();
        wait_for_click();
    }
    /* update client */
    net_write_empty_msg( client, MSG_UNLINK );
    net_write_empty_msg( client, MSG_GAME_EXITED );
    client_prepare_update(); /* clear propably old data as 
                                MSG_GAME_EXITED will cause an update */
    gui_draw(); 
    fade_screen( FADE_IN, FADE_DEF_TIME );
#endif
}

/*
====================================================================
Publics
====================================================================
*/

/*
====================================================================
Create/delete client stuff
====================================================================
*/
int client_create()
{
#ifdef WITH_NETWORK
    int i, j;
    /* assemble */
    cr_bkgnd = create_surf( sdl.screen->w, sdl.screen->h, SDL_SWSURFACE );
    for ( i = 0; i < cr_bkgnd->w; i += cr_wallpaper->w )
        for ( j = 0; j < cr_bkgnd->h; j += cr_wallpaper->h ) {
            DEST( cr_bkgnd, i, j, cr_wallpaper->w, cr_wallpaper->h );
            SOURCE( cr_wallpaper, 0, 0 );
            blit_surf();
        }
    SDL_SetColorKey( cr_bkgnd, 0, 0 );
    /* add 3d frames */
    draw_3dframe( cr_bkgnd, gui->channels->region.x, gui->channels->region.y, gui->channels->region.w, gui->channels->region.h, 2 );
    draw_3dframe( cr_bkgnd, gui->users->region.x, gui->users->region.y, gui->users->region.w, gui->users->region.h, 2 );
    draw_3dframe( cr_bkgnd, gui->levels->region.x, gui->levels->region.y, gui->levels->region.w, gui->levels->region.h, 2 );
    draw_3dframe( cr_bkgnd, gui->games->region.x, gui->games->region.y, gui->games->region.w, gui->games->region.h, 2 );
    draw_3dframe( cr_bkgnd, gui->chat_win->region.x, gui->chat_win->region.y, gui->chat_win->region.w, gui->chat_win->region.h, 2 );
    draw_3dframe( cr_bkgnd, gui->chat_edit->region.x, gui->chat_edit->region.y, gui->chat_edit->region.w, gui->chat_edit->region.h, 2 );
    draw_3dframe( cr_bkgnd, gui->mp_menu->x - 2, gui->mp_menu->y - 2, gui->mp_menu->w + 4, gui->mp_menu->h + 4, 2 );
    menu_set_bkgnd( gui->mp_menu, cr_bkgnd );
    /* initialize network functions */	
    net_init();
    /* create messages */
    net_set_msg( MSG_ERROR, "z" );
    net_set_msg( MSG_SERVER_INFO,  "z" );
    net_set_msg( MSG_IDENTIFY, "bz" );
    net_set_msg( MSG_LOGIN_OKAY, "szz" );
    net_set_msg( MSG_CHAT, "z" );
    net_set_msg( MSG_ADD_USER, "sz" );
    net_set_msg( MSG_REMOVE_USER, "s" );
    net_set_msg( MSG_ADD_USERS, "br" );
    net_set_msg( MSG_ADD_GAME, "szz" );
    net_set_msg( MSG_REMOVE_GAME, "s" );
    net_set_msg( MSG_UPDATE_GAME, "sb" );
    net_set_msg( MSG_ADD_GAMES, "br" );
    net_set_msg( MSG_ADD_CHANNEL, "sz" );
    net_set_msg( MSG_REMOVE_CHANNEL, "s" );
    net_set_msg( MSG_ADD_CHANNELS, "br" );
    net_set_msg( MSG_CHALLENGE, "sszbbbb" );
    net_set_msg( MSG_IP_ADDRESS, "zi" );
    net_set_msg( MSG_LEVEL_COUNT, "b" );
    net_set_msg( MSG_LEVEL_DATA, "zzrr" );
    net_set_msg( MSG_BENCHMARK, "s" );
    net_set_msg( MSG_WINNER, "b" );
    net_set_msg( MSG_DM_HOSTDATA, "bbsr" );
    net_set_msg( MSG_DM_CLIENTDATA, "s" );
    net_set_msg( MSG_MP_HOSTDATA, "iisiirrr" );
    net_set_msg( MSG_MP_CLIENTDATA, "s" );
    net_set_msg( MSG_ROUND_OVER, "iiiiiiiiiiiiiiiiii" );
    net_set_msg( MSG_SWITCH_CHANNEL, "s" );
    net_set_msg( MSG_SET_CHANNEL, "s" );
    net_set_msg( MSG_INIT_TRANSFER, "zzzs" );
    
    /* users, channels, games */
    channels = list_create( LIST_AUTO_DELETE, 0 );
    users = list_create( LIST_AUTO_DELETE, 0 );
    games = list_create( LIST_AUTO_DELETE, 0 );
    return 1;
#else
    return 1;
#endif
}
void client_delete()
{
#ifdef WITH_NETWORK
    free_surf( &cr_bkgnd );
    if ( channels ) list_delete( channels ); channels = 0;
    if ( users ) list_delete( users ); users = 0;
    if ( games ) list_delete( games ); games = 0;
#endif
}

/*
====================================================================
Connect to game server
====================================================================
*/
int client_connect()
{
#ifdef WITH_NETWORK
    char host[16];
    Net_Msg msg;
    if ( strcmp( config.host, "localhost" ) )
        strcpy( host, config.host );
    else
        strcpy( host, "127.0.0.1" );
    if ( ( client = net_connect( host, atoi( config.port ) ) ) ) {
        /* send identification */
        net_init_msg( &msg, MSG_IDENTIFY );
        if ( !net_pack_int8( &msg, PROTOCOL ) ||
             !net_pack_string( &msg, config.username ) || 
             !net_write_msg( client, &msg ) ) {
            sprintf( client_error, "%s\n", net_get_error() );
            return 0;
        }
        else
            if ( net_read_msg( client, &msg, 8000 ) )
                switch ( msg.type ) {
                    case MSG_LOGIN_OKAY:
                        net_unpack_int16( &msg, &client_id );
                        net_unpack_string( &msg, client_name, 16 );
                        net_unpack_string( &msg, client_ip, NET_IP_SIZE );
                        client_prepare_update();
                        return 1;
                    case MSG_ERROR:
                        net_unpack_string( &msg, client_error, 127 );
                        return 0;
                    case MSG_DISCONNECT:
                        strcpy( client_error, "Server has disconnected" );
                        return 0;
                }
        strcpy( client_error, "No response from server" );
        return 0;
    }
    strcpy( client_error, net_get_error() );
    return 0;
#else
    return 1;
#endif
}
void client_disconnect()
{
#ifdef WITH_NETWORK
    if ( client ) {
        net_write_empty_msg( client, MSG_DISCONNECT );
        net_close( client );
    }
#endif
}

/*
====================================================================
Get client error.
====================================================================
*/
char* client_get_error()
{
    if ( client_error[0] != 0 )
        return client_error;
    return 0;
}

/*
====================================================================
Run the chatroom.
====================================================================
*/
int client_run()
{
#ifdef WITH_NETWORK
    int ms;
    char buf[NET_MSG_DATA_LIMIT + 1];
    char name[16];
    Net_Msg msg;
    int num, win, length, remote_version, remote_update;
    int leave = 0, ret = 1, id;
    SDL_Event event;
    CUser *user;
    CChannel *channel;
    Level *level = 0;
    
    client_state = CSTATE_NONE;
    client_error[0] = 0;
    gui_update_levelsets();
    list_clear( gui->chat_win->entries ); gui->chat_win->cur_item = 0;
    list_clear( gui->pause_win->entries ); gui->pause_win->cur_item = 0;
    gui_draw(); refresh_screen( 0, 0, 0, 0 );
    reset_timer();
    group_get_bkgnd( gui->confirm );
    while ( !leave ) {
        group_draw_bkgnd( gui->confirm );
        menu_hide( gui->mp_menu );
        hint_hide();
        if ( event_poll( &event ) )
            if ( gui_handle_event( &event, &id ) )
                switch ( id ) {
                    case ID_OK:
                        group_hide( gui->confirm, 1 );
                        switch ( client_state ) {
                            case CSTATE_CONFIRM_TRANSFER:
                                if ( !net_write_empty_msg( client, MSG_ACCEPT_SET ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                else {
                                    /* prepare to receive levelcount levels */
                                    levels_create_empty_set( mp_level_count );
                                    list_reset( levels ); level = list_next( levels );
                                    client_popup_info( "Receiving levels..." );
                                    client_state = CSTATE_RECEIVE;
                                }
                                break;
                            case CSTATE_ANSWER:
                                if ( !net_write_empty_msg( client, MSG_ACCEPT_CHALLENGE ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                else
                                    client_run_game( 0 );
                                client_state = CSTATE_NONE;
                                break;
                        }
                        break;
                    case ID_CANCEL:
                        group_hide( gui->confirm, 1 );
                        switch ( client_state ) {
                            case CSTATE_RECEIVE:
                                if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                break;
                            case CSTATE_CONFIRM_TRANSFER:
                                if ( !net_write_empty_msg( client, MSG_REJECT_SET ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                break;
                            case CSTATE_AWAIT_TRANSFER_CONFIRMATION:
                                if ( !net_write_empty_msg( client, MSG_CANCEL_TRANSFER ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                break;
                            case CSTATE_AWAIT_ANSWER:
                                if ( !net_write_empty_msg( client, MSG_CANCEL_CHALLENGE ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                break;
                            case CSTATE_ANSWER:
                                if ( !net_write_empty_msg( client, MSG_REJECT_CHALLENGE ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                break;
                        }
                        client_state = CSTATE_NONE;
                        break;
                    case ID_QUIT: 
                        client_error[0] = 0;
                        leave = 1; 
                        break;
                    case ID_CHAT_EDIT:
                        if ( gui->chat_edit->text[0] == 0 ) break; /*do not send empty messages */
                        /* check if we want to transfer a levelset */
                        if ( client_check_transfer( gui->chat_edit->text ) )
                            break;
                        /* send message to all users */
                        if ( !net_build_msg( &msg, MSG_CHAT, gui->chat_edit->text ) ||
                             !net_write_msg( client, &msg ) ) 
                            fprintf( stderr, "%s\n", net_get_error() );
                        /* clear chat edit */
                        gui->chat_edit->text[0] = 0;
                        gui->chat_edit->length = 0;
                        gui_edit_draw( gui->chat_edit );
                        break;
                    case ID_CHALLENGE:
                        if ( client_state != CSTATE_NONE ) break;
                        if ( !gui->users->cur_item ) {
                            client_popup_info( "You must select a user for a challenge." );
                            break;
                        }
                        if ( !gui->levels->cur_item ) {
                            client_popup_info( "You must select the levelset you want to play." );
                            break;    
                        }
                        if ( STRCMP( gui->users->cur_item, client_name ) ) {
                            client_popup_info( "You can't challenge yourself, man!" );
                            break;
                        }
                        if ( ( mp_peer_id = client_get_user_id( gui->users->cur_item ) ) == 0 ) 
                            break;
                        strcpy( mp_peer_name, gui->users->cur_item );
                        strcpy( mp_levelset, gui->levels->cur_item );
                        mp_diff = config.mp_diff;
                        mp_rounds = config.mp_rounds;
                        mp_balls = config.mp_balls;
                        mp_frags = config.mp_frags;
                        /* challenger, challenged, levelset, diff, rounds, frags, balls */
                        net_init_msg( &msg, MSG_CHALLENGE );
                        if ( !net_pack_int16( &msg, client_id ) ||
                             !net_pack_int16( &msg, mp_peer_id ) ||
                             !net_pack_string( &msg, mp_levelset ) ||
                             !net_pack_int8( &msg, mp_diff ) ||
                             !net_pack_int8( &msg, mp_rounds ) ||
                             !net_pack_int8( &msg, mp_frags ) ||
                             !net_pack_int8( &msg, mp_balls ) ||
                             !net_write_msg( client, &msg ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                        else
                            client_popup( CSTATE_AWAIT_ANSWER );
                        break;
                    case ID_CHANNELS:
                        /* switch to selected channel */
                        if ( gui->channels->cur_item )
                            if ( ( channel = client_find_channel_by_name( gui->channels->cur_item ) ) ) {
                                if ( !net_build_msg( &msg, MSG_SWITCH_CHANNEL, channel->id ) ||
                                     !net_write_msg( client, &msg ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                            }
                        break;
                }
        if ( net_read_msg( client, &msg, 0 ) )
            switch ( msg.type ) {
                case MSG_PREPARE_UPDATE:
                    /* clear all data to receive new data
                       the server is about to send */
                    client_prepare_update();
                    break;
                case MSG_CHAT:
                    /* add chatter to chat window */
                    if ( net_unpack_string( &msg, buf, CHAT_MSG_LIMIT + 16 ) )
                        gui_add_chat_msg( gui->chat_win, CHAT_WIN_LINE_WIDTH, buf, 0 );
                    break;
                case MSG_ADD_CHANNEL:
                    /* add a channel to list */
                    if ( net_unpack_int16( &msg, &num ) )
                    if ( net_unpack_string( &msg, name, 16 ) ) {
                        client_add_channel( num, name );
                        gui_list_draw( gui->channels );
                    }
                    break;
                case MSG_REMOVE_CHANNEL:
                    /* remove channel from list */
                    if ( net_unpack_int16( &msg, &num ) ) {
                        client_remove_channel( num );
                        gui_list_draw( gui->channels );
                    }
                    break;
                case MSG_SET_CHANNEL:
                    /* select channel by this id -- only update the channel list */
                    if ( net_unpack_int16( &msg, &num ) ) {
                        gui->channels->cur_item = client_find_channel_name( num );
                        gui_list_draw( gui->channels );
                    }
                    break;
                case MSG_ADD_USER:
                    /* add a user to users list */
                    if ( net_unpack_int16( &msg, &num ) )
                        if ( net_unpack_string( &msg, name, 16 ) ) {
                            client_add_user( num, name );
                            gui_list_draw( gui->users );
                        }
                    break;
                case MSG_REMOVE_USER:
                    /* remove a user */
                    if ( net_unpack_int16( &msg, &num ) ) {
                        client_remove_user( num );
                        gui_list_draw( gui->users );
                    }
                    break;
                case MSG_ADD_GAME:
                    if ( net_unpack_int16( &msg, &num ) )
                    if ( net_unpack_string( &msg, buf, 48 ) )
                    if ( net_unpack_string( &msg, name, 16 ) ) {
                        client_add_game( num, buf, name );
                        gui_list_draw( gui->games );
                    }
                    break;
                case MSG_REMOVE_GAME:
                    if ( net_unpack_int16( &msg, &num ) ) {
                        client_remove_game( num );
                        gui_list_draw( gui->games );
                    }
                    break;
                case MSG_UPDATE_GAME:
                    if ( net_unpack_int16( &msg, &num ) )
                    if ( net_unpack_int8( &msg, &win ) ) {
                        client_update_game( num, win  );
                        gui_list_draw( gui->games );
                    }
                    break;
                case MSG_ERROR:
                    /* error message */
                    if ( net_unpack_string( &msg, client_error, 127 ) ) {
                        fprintf( stderr, "%s\n", client_error );
                        gui_add_chat_msg( gui->chat_win, CHAT_WIN_LINE_WIDTH, client_error, 1 );
                    }
                    break;
                case MSG_DISCONNECT:
                    /* disconnect */
                    leave = 1; ret = 0;
                    break;
                case MSG_CHALLENGE:
                    if ( client_state == CSTATE_ANSWER || client_state == CSTATE_AWAIT_ANSWER ) break;
                    if ( net_unpack_int16( &msg, &mp_peer_id ) ) /* challenger */
                    if ( net_unpack_int16( &msg, &num ) ) /* this client (challenged) */
                    if ( net_unpack_string( &msg, mp_levelset, 16 ) )
                    if ( net_unpack_int8( &msg, &mp_diff ) )
                    if ( net_unpack_int8( &msg, &mp_rounds ) )
                    if ( net_unpack_int8( &msg, &mp_frags ) )
                    if ( net_unpack_int8( &msg, &mp_balls ) )
                    if ( ( user = client_find_user( mp_peer_id ) ) ) {
                        strcpy_lt( mp_peer_name, user->name, 15 );
                        client_popup( CSTATE_ANSWER );
                    }
                    break;
                case MSG_SERVER_INFO:
                    if ( net_unpack_string( &msg, buf, NET_MSG_DATA_LIMIT + 1 ) )
                        gui_add_chat_msg( gui->chat_win, CHAT_WIN_LINE_WIDTH, buf, 1 );
                    break;
                case MSG_BUSY:
                    if ( client_state == CSTATE_AWAIT_ANSWER ) {
                        sprintf( buf, "%s is busy at the moment.", mp_peer_name );
                        client_popup_info( buf );
                    }
                    break;
                case MSG_REJECT_CHALLENGE:
                    if ( client_state == CSTATE_AWAIT_ANSWER ) {
                        sprintf( buf, "%s is too scared to accept your challenge.", mp_peer_name );
                        client_popup_info( buf );
                        if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                    }
                    break;
                case MSG_CANCEL_CHALLENGE:
                    if ( client_state == CSTATE_ANSWER ) {
                        sprintf( buf, "%s got cold feet.", mp_peer_name );
                        client_popup_info( buf );
                        if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                    }
                    break;
                case MSG_ACCEPT_CHALLENGE:
                    if ( client_state == CSTATE_AWAIT_ANSWER ) {
                        group_hide( gui->confirm, 1 );
                        client_run_game( 1 );
                        client_state = CSTATE_NONE;
                    }
                    break;
               case MSG_INIT_TRANSFER:
                   if ( net_unpack_string( &msg, mp_peer_name, 16 ) ) /* sender */
                   if ( net_unpack_string( &msg, mp_levelset, 16 ) ) /* set name */
                   if ( net_unpack_string( &msg, mp_levelset_version, 16 ) ) /* version */
                   if ( net_unpack_int16( &msg, &mp_level_count ) ) { /* level count */
                       /* test if we got a newer version of this set */
                       sprintf( name, "~%s", mp_levelset );
                       if ( levels_load( name ) ) {
                           parse_version( mp_levelset_version, &remote_version, &remote_update );
                           if ( levelset_version > remote_version ||
                                ( levelset_version == remote_version &&
                                  levelset_update >= remote_update ) ) {
                               net_write_empty_msg( client, MSG_OLDER_SET );
                               if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                                   fprintf( stderr, "%s\n", net_get_error() );
                               break;
                           }
                       }
                       client_popup( CSTATE_CONFIRM_TRANSFER );
                   }
                   break;
               case MSG_CANCEL_TRANSFER:
                    if ( client_state == CSTATE_CONFIRM_TRANSFER ) {
                        sprintf( buf, "%s cancelled the transfer.", mp_peer_name );
                        client_popup_info( buf );
                        if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                    }
                   break;
               case MSG_OLDER_SET:
                    if ( client_state == CSTATE_AWAIT_TRANSFER_CONFIRMATION ) {
                        sprintf( buf, "%s's levelset is up-to-date.", mp_peer_name );
                        client_popup_info( buf );
                        if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                    }
                   break;
               case MSG_REJECT_SET:
                    if ( client_state == CSTATE_AWAIT_TRANSFER_CONFIRMATION ) {
                        sprintf( buf, "%s rejected your levelset.", mp_peer_name );
                        client_popup_info( buf );
                        if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                    }
                   break;
               case MSG_ACCEPT_SET:
                    if ( client_state == CSTATE_AWAIT_TRANSFER_CONFIRMATION ) {
                        /* send all levels */
                        list_reset( levels );
                        while ( ( level = list_next( levels ) ) ) {
                            net_init_msg( &msg, MSG_LEVEL_DATA );
                            if ( !net_pack_string( &msg, level->author ) ||
                                 !net_pack_string( &msg, level->name ) ||
                                 !net_pack_raw( &msg, level->bricks, sizeof( level->bricks ) ) ||
                                 !net_pack_raw( &msg, level->extras, sizeof( level->extras ) ) ||
                                 !net_write_msg( client, &msg ) ) {
                                fprintf( stderr, "%s\n", net_get_error() );
                                break;
                            }
                        }
                        sprintf( buf, "Levelset send." );
                        client_popup_info( buf );
                        if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                            fprintf( stderr, "%s\n", net_get_error() );
                    }
                   break;
                case MSG_LEVEL_DATA:
                    if ( level == 0 ) break;
                    if ( client_state == CSTATE_RECEIVE ) {
                        if ( !net_unpack_string( &msg, level->author, 31 ) ||
                             !net_unpack_string( &msg, level->name, 31 ) ||
                             !net_unpack_raw( &msg, level->bricks, &length, sizeof( level->bricks ) ) ||
                             !net_unpack_raw( &msg, level->extras, &length, sizeof( level->extras ) ) ) {
                            fprintf( stderr, "%s\n", net_get_error() );
                            client_state = CSTATE_NONE;
                            group_hide( gui->confirm, 1 );
                            break;
                        }
                        else {
                            if ( ( level = list_next( levels ) ) == 0 ) {
                                /* save to disk and unlink connection */
                                parse_version( mp_levelset_version, 
                                               &levelset_version, &levelset_update );
                                levels_save( mp_levelset );
                                if ( !net_write_empty_msg( client, MSG_UNLINK ) )
                                    fprintf( stderr, "%s\n", net_get_error() );
                                group_hide( gui->confirm, 1 );
                                client_state = CSTATE_NONE;
                                gui->levels->cur_item = 0;
                                levelsets_load_names();
                                gui_update_levelsets(); 
                                gui_list_draw( gui->levels );
                            }
                        }
                    }
                    break;
            }
        ms = get_time();
        menu_update( gui->mp_menu, ms );
        hint_update( ms );
        menu_show( gui->mp_menu );
        gui_list_draw( gui->levels );
        group_get_bkgnd( gui->confirm );
        group_draw( gui->confirm );
        hint_show();
        refresh_rects();
        SDL_Delay( 10 );
        if ( client->lost ) {
            ret = 0; leave = 1;
            sprintf( client_error, "Connection lost." );
        }
    }
    return ret;
#else
    return 1;
#endif
}
