/***************************************************************************
                          frame.c  -  description
                             -------------------
    begin                : Fri Sep 7 2001
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lbreakout.h"
#include "config.h"
#include "levels.h"
#include "player.h"
#include "display.h"
#include "paddle.h"
#include "bricks.h"
#include "extras.h"
#include "frame.h"
#ifdef SOUND
#include "audio.h"
#endif

extern Sdl sdl;
extern SDL_Surface *offscreen;
extern SDL_Surface *bkgnd; /* background + frame */
extern SDL_Surface *brick_pic; /* brick graphics in a horizontal order */
/* parts of the frame: left, top, right */
extern SDL_Surface *frame_left, *frame_top, *frame_right;
extern SDL_Surface *frame_left_shadow, *frame_top_shadow, *frame_right_shadow;
extern SDL_Surface *frame_mp_left, *frame_mp_right;
extern SDL_Surface *frame_mp_left_shadow, *frame_mp_right_shadow;
extern SDL_Surface *lamps; /* life lamps */
float score = 0; /* current score */
int dest_score; /* final score */
int score_x_offset = 54; /* display score at this position (right align)*/
int score_x, score_y = BRICK_HEIGHT / 2 - 2; /* center position where to write score */
int new_life_y; /* if new life lamp is drawn do this at this y position */
int name_x = 404, name_y = BRICK_HEIGHT / 2 - 2; /* offset (right side of screen) */
extern int shadow_size;
extern Config config;
int info_x;
int info_y; /* absolute position is info_y + info_offset */
int info_offsets[EX_NUMBER]; /* offset at which extra info is displayed */
int paddle_info_y[4];
int paddle_info_offsets[EX_NUMBER]; /* specialized extras */
extern int active[EX_NUMBER];
extern int extra_time[EX_NUMBER];
extern SDL_Surface *extra_pic;
#ifdef SOUND
extern Sound_Chunk *wav_life_up, *wav_life_down;
#endif
extern int game_type;
extern Paddle *paddles[4];
extern int paddle_count;
extern Font *display_font;
extern int warp_ok; /* warp_limit/warp_ok is defined back at bricks.c */
int warp_blink; 
Delay warp_delay;
int warp_x, warp_y;
extern SDL_Surface *warp_pic;

/*
====================================================================
Locals
====================================================================
*/

/*
====================================================================
Publics
====================================================================
*/

/*
====================================================================
Create frame by loading and assembling all graphics and loading
additional graphics.
====================================================================
*/
void frame_create()
{
    delay_set( &warp_delay, 500 );
    warp_x = sdl.screen->w - BRICK_WIDTH + ( BRICK_WIDTH - warp_pic->w ) / 2;
    warp_y = sdl.screen->h - warp_pic->h;
    warp_blink = 0;
}

/*
====================================================================
Inititate the frame to the according game type.
Set the position of the extras displayed, enable life lamps etc.
====================================================================
*/
void frame_init()
{
    /* setup position of extra info */
    info_x = sdl.screen->w - BRICK_WIDTH;
    if ( game_type == GT_NETWORK ) {
        /* general */
        info_y = 178;
        info_offsets[EX_METAL] = 26;
        info_offsets[EX_WEAK_BALL] = 26;
        info_offsets[EX_EXPL_BALL] = 26;
        info_offsets[EX_FAST] = 52;
        info_offsets[EX_SLOW] = 52;
        info_offsets[EX_CHAOS] = 78;
        /* paddles */
        paddle_info_y[0] = 271;
        paddle_info_y[1] = 7;
        paddle_info_offsets[EX_SLIME] = 26;
        paddle_info_offsets[EX_WEAPON] = 52;
        paddle_info_offsets[EX_WALL] = 78;
        paddle_info_offsets[EX_GOLDSHOWER] = 104;
        paddle_info_offsets[EX_GHOST_PADDLE] = 130;
        paddle_info_offsets[EX_BONUS_MAGNET] = 156;
        paddle_info_offsets[EX_MALUS_MAGNET] = 156;
    }
    else {
        /* in single player we ignore the second paddle
           and model general and paddle0 extras so that
           they result in the old layout */
        info_y = 48; paddle_info_y[0] = 48;
        paddle_info_offsets[EX_GOLDSHOWER] = 30;
        paddle_info_offsets[EX_SLIME] = 60;
        info_offsets[EX_METAL] = 90;
        info_offsets[EX_WEAK_BALL] = 90;
        info_offsets[EX_EXPL_BALL] = 90;
        paddle_info_offsets[EX_WALL] = 120;
        paddle_info_offsets[EX_WEAPON] = 150;
        info_offsets[EX_FAST] = 180;
        info_offsets[EX_SLOW] = 180;
        info_offsets[EX_CHAOS] = 210;
        paddle_info_offsets[EX_GHOST_PADDLE] = 240;
        paddle_info_offsets[EX_BONUS_MAGNET] = 270;
        paddle_info_offsets[EX_MALUS_MAGNET] = 270;
    }
}

/*
====================================================================
Free all resources created by frame_create()
====================================================================
*/
void frame_delete()
{
}
/*
====================================================================
Draw frame to offscreen and to bkgnd as frame won't change while
playing.
====================================================================
*/
void frame_draw()
{
    SDL_Surface *fr_left = (game_type==GT_LOCAL)?frame_left:frame_mp_left;
    SDL_Surface *fr_right = (game_type==GT_LOCAL)?frame_right:frame_mp_right;
    SDL_Surface *fr_left_shadow = (game_type==GT_LOCAL)?frame_left_shadow:frame_mp_left_shadow;
    SDL_Surface *fr_right_shadow = (game_type==GT_LOCAL)?frame_right_shadow:frame_mp_right_shadow;
    /* left and right part are always drawn */
    /* left part */
    if ( config.shadow ) {
        DEST( offscreen, shadow_size, shadow_size, fr_left_shadow->w, fr_left_shadow->h );
        SOURCE( fr_left_shadow, 0, 0 );
        alpha_blit_surf( SHADOW_ALPHA );
    }
    DEST( offscreen, 0, 0, fr_left->w, fr_left->h );
    SOURCE( fr_left, 0, 0 );
    blit_surf();
    if ( config.shadow ) {
        DEST( bkgnd, shadow_size, shadow_size, fr_left_shadow->w, fr_left_shadow->h );
        SOURCE( fr_left_shadow, 0, 0 );
        alpha_blit_surf( SHADOW_ALPHA );
    }
    DEST( bkgnd, 0, 0, fr_left->w, fr_left->h );
    SOURCE( fr_left, 0, 0 );
    blit_surf();
    if ( game_type == GT_LOCAL ) {
        /* add top */
        if ( config.shadow ) {
            DEST( offscreen, fr_left->w + shadow_size, shadow_size, 
                  frame_top_shadow->w, frame_top_shadow->h );
            SOURCE( frame_top_shadow, 0, 0 );
            alpha_blit_surf( SHADOW_ALPHA );
        }
        DEST( offscreen, fr_left->w, 0, frame_top->w, frame_top->h );
        SOURCE( frame_top, 0, 0 );
        blit_surf();
        if ( config.shadow ) {
            DEST( bkgnd, fr_left->w + shadow_size, shadow_size, 
                  frame_top_shadow->w, frame_top_shadow->h );
            SOURCE( frame_top_shadow, 0, 0 );
            alpha_blit_surf( SHADOW_ALPHA );
        }
        DEST( bkgnd, fr_left->w, 0, frame_top->w, frame_top->h );
        SOURCE( frame_top, 0, 0 );
        blit_surf();
    }
    /* right part */
    if ( config.shadow ) {
        DEST( offscreen, sdl.screen->w - fr_right_shadow->w + shadow_size, shadow_size,
              fr_right_shadow->w, fr_right_shadow->h );
        SOURCE( fr_right_shadow, 0, 0 );
        alpha_blit_surf( SHADOW_ALPHA );
    }
    DEST( offscreen, sdl.screen->w - fr_right->w, 0, fr_right->w, fr_right->h );
    SOURCE( fr_right, 0, 0 );
    blit_surf();
    if ( config.shadow ) {
        DEST( bkgnd, sdl.screen->w - fr_right_shadow->w + shadow_size, shadow_size,
              fr_right_shadow->w, fr_right_shadow->h );
        SOURCE( fr_right_shadow, 0, 0 );
        alpha_blit_surf( SHADOW_ALPHA );
    }
    DEST( bkgnd, sdl.screen->w - fr_right->w, 0, fr_right->w, fr_right->h );
    SOURCE( fr_right, 0, 0 );
    blit_surf();
}
/*
====================================================================
Add life lamps at left side of frame in offscreen
====================================================================
*/
void frame_draw_lives( int lives, int max_lives )
{
    int i;
    /* substract one life to have the same result like in old LBreakout */
    /* at maximum ten lamps may be displayed */
    for ( i = 0; i < 10; i++ ) {
        DEST( offscreen, 0, ( MAP_HEIGHT - i - 1 ) * BRICK_HEIGHT, BRICK_WIDTH, BRICK_HEIGHT );
        if ( i < lives - 1 )
            SOURCE( lamps, 0, BRICK_HEIGHT )
        else
            if ( i < max_lives - 1 )
                SOURCE( lamps, 0, 0 )
            else
                SOURCE( lamps, 0, BRICK_HEIGHT * 2 )
        blit_surf();
    }
    /* get position of next lamp */
    new_life_y = sdl.screen->h - lives * BRICK_HEIGHT;
}
/*
====================================================================
Add one new life at offscreen and screen (as this will happen in
game )
====================================================================
*/
void frame_add_life()
{
	if ( !active[EX_DARKNESS] ) {
    	DEST( sdl.screen, 0, new_life_y, BRICK_WIDTH, BRICK_HEIGHT );
    	SOURCE( lamps, 0, BRICK_HEIGHT );
    	blit_surf();
    	add_refresh_rect( 0, new_life_y, BRICK_WIDTH, BRICK_HEIGHT );
	}		
    new_life_y -= BRICK_HEIGHT;
#ifdef SOUND
    sound_play( wav_life_up );
#endif
}
/*
====================================================================
Switch of a life lamp
====================================================================
*/
void frame_remove_life()
{
    new_life_y += BRICK_HEIGHT;
    if ( new_life_y >= sdl.screen->h ) return;
	if ( !active[EX_DARKNESS] ) {
	    DEST( sdl.screen, 0, new_life_y, BRICK_WIDTH, BRICK_HEIGHT );
	    SOURCE( lamps, 0, 0 );
	    blit_surf();
	    add_refresh_rect( 0, new_life_y, BRICK_WIDTH, BRICK_HEIGHT );
	}		
#ifdef SOUND
    sound_play( wav_life_down );
#endif
}
/*
====================================================================
Display extra information on right side of screen.
====================================================================
*/
void frame_info_hide()
{
    int i, j;
    if ( !config.bonus_info ) return;
    for ( i = 0; i < EX_NUMBER; i++ )
        if ( info_offsets[i] > 0 && active[i] ) {
            DEST( sdl.screen, info_x, info_y + info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
            SOURCE( offscreen, info_x, info_y + info_offsets[i] );
            blit_surf();
            add_refresh_rect( info_x, info_y + info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
        }
    for ( i = 0; i < EX_NUMBER; i++ )
        for ( j = 0; j < paddle_count; j++ )
            if ( paddle_info_offsets[i] > 0 && paddles[j]->active[i] ) {
                DEST( sdl.screen, info_x, paddle_info_y[j] + paddle_info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
                SOURCE( offscreen, info_x, paddle_info_y[j] + paddle_info_offsets[i] );
                blit_surf();
                add_refresh_rect( info_x, paddle_info_y[j] + paddle_info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
            }
}
void frame_info_show()
{
    char str[12];
    int i, j;
    if ( !config.bonus_info ) return;
    display_font->align = ALIGN_X_CENTER | ALIGN_Y_CENTER;
    for ( i = 0; i < EX_NUMBER; i++ ) {
        if ( info_offsets[i] > 0 && active[i] ) {
            /* picture */
            DEST( sdl.screen, info_x, info_y + info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
            fill_surf( 0x0 );
            SOURCE( extra_pic, i * BRICK_WIDTH, 0 );
            alpha_blit_surf( 96 );
            /* remaining extra_time */
            sprintf(str, "%i", (extra_time[i] / 1000) + 1);
            /* write text without saving last rect as this is used for score update */
            display_font->save_last = 0;
            write_text( display_font, sdl.screen,
                        info_x + ( BRICK_WIDTH >> 1 ), info_y + info_offsets[i] + ( BRICK_HEIGHT >> 1 ),
                        str, OPAQUE );
            display_font->save_last = 1;
            /* refresh */
            add_refresh_rect( info_x, info_y + info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
        }
	}
    for ( i = 0; i < EX_NUMBER; i++ )
        for ( j = 0; j < paddle_count; j++ )
            if ( paddle_info_offsets[i] > 0 && paddles[j]->active[i] ) {
                if ( i == EX_WALL && paddles[j]->extra_time[i] <= 0 ) continue;
                /* picture */
                DEST( sdl.screen, info_x, paddle_info_y[j] + paddle_info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
                fill_surf( 0x0 );
                SOURCE( extra_pic, i * BRICK_WIDTH, 0 );
                alpha_blit_surf( 128 );
                /* remaining extra_time */
                sprintf(str, "%i", (paddles[j]->extra_time[i] / 1000) + 1);
                /* write text without saving last rect as this is used for score update */
                display_font->save_last = 0;
                write_text( display_font, sdl.screen,
                            info_x + ( BRICK_WIDTH >> 1 ), paddle_info_y[j] + paddle_info_offsets[i] + ( BRICK_HEIGHT >> 1 ),
                            str, OPAQUE );
                display_font->save_last = 1;
                /* refresh */
                add_refresh_rect( info_x, paddle_info_y[j] + paddle_info_offsets[i], BRICK_WIDTH, BRICK_HEIGHT );
            }
}

/*
====================================================================
Blink the warp icon.
====================================================================
*/
void frame_warp_icon_hide()
{
    if ( !warp_ok ) return;
    DEST( sdl.screen, warp_x, warp_y, warp_pic->w, warp_pic->h );
    SOURCE( offscreen, warp_x, warp_y );
    blit_surf();
    add_refresh_rect( warp_x, warp_y, warp_pic->w, warp_pic->h );
}
void frame_warp_icon_show()
{
    if ( !warp_ok ) return;
    if ( !warp_blink ) return;
    DEST( sdl.screen, warp_x, warp_y, warp_pic->w, warp_pic->h );
    SOURCE( warp_pic, 0, 0 );
    if ( active[EX_DARKNESS] )
        alpha_blit_surf( 128 );
    else
        blit_surf();
}
void frame_warp_icon_update( int ms )
{
    if ( !warp_ok ) return;
    if ( delay_timed_out( &warp_delay, ms ) )
        warp_blink = !warp_blink;
}
