/***************************************************************************
                          gui.c  -  description
                             -------------------
    begin                : Sun Apr 28 12:02:57 CEST 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lbreakout.h"
#include "config.h"
#include "../common/net.h"
#include "manager.h"
#include "windows.h"
#include "gui.h"
#include "levels.h"

#define INRECT( mx, my, rect ) ( (mx>=rect.x)&&(mx<rect.x+rect.w)&&(my>=rect.y)&&(my<rect.y+rect.h) )

#define HINT_ 0
#define HINT_DIFF "The diffculty influences ball speed, paddle size and score."
#define HINT_ROUNDS "This is the number of rounds played per level. Winning a round scores one point and the player with the most points wins the match."
#define HINT_FRAGS "In a deathmatch level (no bricks) a player gains a frag when the opponent looses a ball. If a player reaches this number of frags he wins the round.#In normal levels (with bricks) this option is ignored and the level is over when all bricks are cleared."
#define HINT_BALLS "In a deathmatch level (no bricks) each player may fire multiple balls up to this number. If a player looses a ball he may fire it again.#In normal levels (with bricks) this option is ignored and a player may bring back only one ball."
#define HINT_GAME_PORT "The actual game is running via a direct connection and the challenger opens his local gameserver at this port.#NOTE: This is an edit thus you have to type in the port. Clicking doesn't work."
#define HINT_CHALLENGE "Challenge the selected user with the current settings."

/*
====================================================================
Create/delete button. The passed image is just a pointer.
====================================================================
*/
extern Sdl sdl;
extern SDL_Surface *cr_up, *cr_down, *cr_track, *cr_quit, *cr_chall, *cr_bkgnd;
extern SDL_Surface *fr_hori, *fr_vert, *fr_luc, *fr_llc, *fr_ruc, *fr_rlc;
extern List *levelset_names;
extern Config config;
extern Font *chat_font_error;
extern Font *chat_font_name;
extern int motion;
extern int motion_x, motion_y;

GUI *gui = 0;
Button *last_button = 0;

/*
====================================================================
Locals
====================================================================
*/

/*
====================================================================
Create a frame surface
====================================================================
*/
SDL_Surface *gui_create_frame( int w, int h )
{
    int i;
    SDL_Surface *surf = create_surf( w, h, SDL_SWSURFACE );
    SDL_FillRect( surf, 0, 0x0 );
    /* upper, lower horizontal beam */
    for ( i = 0; i < w; i += fr_hori->w ) {
        DEST( surf, i, 0, fr_hori->w, fr_hori->h );
        SOURCE( fr_hori, 0, 0 );
        blit_surf();
        DEST( surf, i, h - fr_hori->h, fr_hori->w, fr_hori->h );
        SOURCE( fr_hori, 0, 0 );
        blit_surf();
    }
    /* left, right vertical beam */
    for ( i = 0; i < h; i += fr_vert->h ) {
        DEST( surf, 0, i, fr_vert->w, fr_vert->h );
        SOURCE( fr_vert, 0, 0 );
        blit_surf();
        DEST( surf, w - fr_vert->w, i, fr_vert->w, fr_vert->h );
        SOURCE( fr_vert, 0, 0 );
        blit_surf();
    }
    /* left upper corner */
    DEST( surf, 0, 0, fr_luc->w, fr_luc->h );
    SOURCE( fr_luc, 0, 0 );
    blit_surf();
    /* left lower corner */
    DEST( surf, 0, h - fr_llc->h, fr_llc->w, fr_llc->h );
    SOURCE( fr_llc, 0, 0 );
    blit_surf();
    /* right upper corner */
    DEST( surf, w - fr_ruc->w, 0, fr_ruc->w, fr_ruc->h );
    SOURCE( fr_ruc, 0, 0 );
    blit_surf();
    /* right lower corner */
    DEST( surf, w - fr_rlc->w, h - fr_rlc->h, fr_rlc->w, fr_rlc->h );
    SOURCE( fr_rlc, 0, 0 );
    blit_surf();
    return surf;
}

/*
====================================================================
Publics
====================================================================
*/

/*
====================================================================
Create/delete button. The passed image is just  pointer.
====================================================================
*/
GUI_Button* gui_button_create( int id, int x, int y, int w, int h, SDL_Surface *image )
{
    GUI_Button *button = calloc( 1, sizeof( GUI_Button ) );
    if ( button == 0 ) {
        fprintf( stderr, "Out of memory" );
        return 0;
    }
    button->image = image;
    button->region.x = x;
    button->region.y = y;
    button->region.w = w;
    button->region.h = h;
    button->id = id;
    return button;
}
void gui_button_delete( GUI_Button *button )
{
    if ( button ) free ( button );
}
/*
====================================================================
Draw and handle button. Handling returns True if button was clicked.
====================================================================
*/
void gui_button_draw( GUI_Button *button )
{
    if ( button->id >= 0 ) {
        DEST( sdl.screen, button->region.x, button->region.y, button->region.w, button->region.h );
        SOURCE( cr_bkgnd, button->region.x, button->region.y );
        blit_surf();
    }
    DEST( sdl.screen, button->region.x, button->region.y, button->region.w, button->region.h );
    SOURCE( button->image, button->state * button->region.w, 0 );
    blit_surf();
    add_refresh_rect( button->region.x, button->region.y, button->region.w, button->region.h );
}
int gui_button_handle( GUI_Button *button, SDL_Event *event, int *id )
{
    if ( event->type == SDL_MOUSEBUTTONUP )
        if ( INRECT( event->button.x, event->button.y, button->region ) ) {
            *id = button->id;
            return 1;
        }
    return 0;
}

/*
====================================================================
Create/delete a list and use default buttons and fonts.
If 'select' is false this is just a memo box.
====================================================================
*/
GUI_List *gui_list_create( int id, int x, int y, int w, int h, int border, int cell_h, int select )
{
    GUI_List *list = calloc( 1, sizeof( GUI_List ) );
    if ( list == 0 ) {
        fprintf( stderr, "Out of memory" );
        return 0;
    }
    list->id = id;
    list->region.x = x;
    list->region.y = y;
    list->region.w = w;
    list->region.h = h;
    list->border = border;
    list->entries = list_create( LIST_AUTO_DELETE, LIST_NO_CALLBACK );
    list->cell_h = cell_h;
    list->length = (h - 2 * border) / cell_h;
    list->scroll_w = list->scroll_h = 24;
    list->track_h = h - 2 * border - 2 * list->scroll_h;
    list->track_y = 0;
    list->track_offset = y + border + list->scroll_h;
    list->up = gui_button_create( -1, x + w - border - list->scroll_w, y + border, list->scroll_w, list->scroll_h, cr_up );
    list->down = gui_button_create( -1, x + w - border - list->scroll_w, y + h - border - list->scroll_h, list->scroll_w, list->scroll_h, cr_down );
    list->track = gui_button_create( -1, x + w - border - list->scroll_w, y + border + list->scroll_h, list->scroll_w, list->scroll_h, cr_track );
    list->font = load_fixed_font( "f_small_yellow.png", 32, 96, 8, SDL_SWSURFACE );
    list->hfont = load_fixed_font( "f_small_white.png", 32, 96, 8, SDL_SWSURFACE );
    list->select = select;
    return list;
}
void gui_list_delete( GUI_List *list )
{
    if ( list ) {
        if ( list->entries ) list_delete( list->entries );
        free_font( &list->font );
        free_font( &list->hfont );
        gui_button_delete( list->up );
        gui_button_delete( list->down );
        gui_button_delete( list->track );
        free( list );
    }
}
/*
====================================================================
Set list items. The old list is deleted and the new one is
used instead.
====================================================================
*/
void gui_list_set_items( GUI_List *list, List *items )
{
    if ( list->entries ) list_delete( list->entries );
    list->entries = items;
}
/*
====================================================================
Add items to list and remove old items if they would exceed limit
(for chatwindow). The passed list is deleted when finished.
====================================================================
*/
void gui_list_add_items( GUI_List *list, List *items, int limit )
{
    char *str;
    list_reset( items );
    while ( ( str = list_next( items ) ) ) {
        if ( list->entries->count >= limit )
            list_delete_pos( list->entries, 0 );
        list_add( list->entries, strdup( str ) );
    }
    list_delete( items );
}
/*
====================================================================
Draw listbox
====================================================================
*/
void gui_list_draw( GUI_List *list )
{
    char *item; char *name_end;
    Font *font;
    int i, j;
    DEST( sdl.screen, list->region.x, list->region.y, list->region.w, list->region.h );
    SOURCE( cr_bkgnd, list->region.x, list->region.y );
    blit_surf();
    for ( i = list->offset, j = 0; j < list->length; i++, j ++ ) {
        item = list_get( list->entries, i );
        if ( item == 0 ) break;
        font = list->font;
        if ( item[0] == '!' ) {
            item++;
            font = chat_font_error;
        }
        font->align = ALIGN_X_LEFT | ALIGN_Y_TOP;
        if ( list->cur_item == item )
            font = list->hfont;
        write_text( font, sdl.screen, list->region.x + list->border, 
                    list->region.y + list->border + j * list->cell_h,
                    item, OPAQUE );
        /* overwrite the name with white */
        if ( item[0] == '<' && ( name_end = strchr( item, '>' ) ) ) {
            name_end[1] = 0;
            write_text( chat_font_name, sdl.screen, list->region.x + list->border, 
                        list->region.y + list->border + j * list->cell_h,
                        item, OPAQUE );
            name_end[1] = ' ';
        }
    }
    gui_button_draw( list->up );
    gui_button_draw( list->down );
    gui_button_draw( list->track );
    add_refresh_rect( list->region.x, list->region.y, list->region.w, list->region.h );
}
/*
====================================================================
Handle event and update the list in screen and add refresh rects.
Return True if selection has changed.
====================================================================
*/
static int gui_list_scroll_up( GUI_List *list )
{
    int old_pos = list->offset;
    list->offset -= list->length / 2;
    if ( list->offset < 0 )
        list->offset = 0;
    if ( list->offset != old_pos )
        return 1;
    return 0;
}
static int gui_list_scroll_down( GUI_List *list )
{
    int old_pos = list->offset;
    list->offset += list->length / 2;
    if ( list->offset > list->entries->count - list->length )
        list->offset = list->entries->count - list->length;
    if ( list->offset < 0 ) list->offset = 0;
    if ( list->offset != old_pos )
        return 1;
    return 0;
}
int gui_list_handle( GUI_List *list, SDL_Event *event, int *id )
{
    char *item;
    int i, ret = 0, redraw = 0;
    SDL_Rect region;
    switch ( event->type ) {
        case SDL_MOUSEBUTTONUP:
            if ( event->button.button == WHEEL_UP ) {
                if ( gui_list_scroll_up( list ) )
                    redraw = 1;
            }
            else
            if ( event->button.button == WHEEL_DOWN ) {
                if ( gui_list_scroll_down( list ) )
                    redraw = 1;
            }
            else 
                if ( list->select ) {
                    region.w = list->region.w - list->scroll_w - ( list->border << 1 );
                    region.h = list->cell_h;
                    region.x = list->region.x + list->border;
                    region.y = list->region.y + list->border;
                    for ( i = 0; i < list->length; i++ ) {
                        if ( INRECT( event->button.x, event->button.y, region ) )
                            if ( ( item = list_get( list->entries, i + list->offset ) ) ) {
                                list->cur_item = item;
                                redraw = 1;
                                ret = 1;
                                break;
                            }
                        region.y += list->cell_h;
                    }
                }
            if ( !ret ) {
                if ( INRECT( event->button.x, event->button.y, list->up->region ) ) {
                    if ( gui_list_scroll_up( list ) )
                        redraw = 1;
                }
                else
                if ( INRECT( event->button.x, event->button.y, list->down->region ) ) {
                    if ( gui_list_scroll_down( list ) )
                        redraw = 1;
                }
            }
            break;
    }
    if ( redraw )
        gui_list_draw( list );
    if ( ret ) *id = list->id;
    return ret;
}

/*
====================================================================
Chat edit
====================================================================
*/
/*
====================================================================
Create/delete edit
====================================================================
*/
GUI_Edit *gui_edit_create( int id, int x, int y, int w, int h, int border, int lw )
{
    GUI_Edit *edit = calloc( 1, sizeof( GUI_Edit ) );
    if ( edit == 0 ) {
        fprintf( stderr, "Out of memory" );
        return 0;
    }
    edit->region.x = x;
    edit->region.y = y;
    edit->region.w = w;
    edit->region.h = h;
    edit->line_w = lw;
    edit->id = id;
    edit->border = border;
    edit->font = load_fixed_font( "f_small_yellow.png", 32, 96, 8, SDL_SWSURFACE );
    return edit;
}
void gui_edit_delete( GUI_Edit *edit )
{
    if ( edit ) {
        free_font( &edit->font );
        free( edit );
    }
}
/*
====================================================================
Draw edit
====================================================================
*/
void gui_edit_draw( GUI_Edit *edit )
{
    int i, j, length, size;
    char str[edit->line_w + 1];
    DEST( sdl.screen, edit->region.x, edit->region.y, edit->region.w, edit->region.h );
    SOURCE( cr_bkgnd, edit->region.x, edit->region.y );
    blit_surf();
    length = strlen( edit->text );
    for ( i = 0, j = 0; i < length; i += edit->line_w, j++ ) {
        memset( str, 0, edit->line_w + 1 );
        size = edit->line_w;
        if ( i + size >= length )
            size = length - i;
        strncpy( str, edit->text + i, size );
        write_text( edit->font, sdl.screen, edit->region.x + edit->border, 
                    edit->region.y + edit->border + j * edit->font->height, str, OPAQUE );
    }
    add_refresh_rect( edit->region.x, edit->region.y, edit->region.w, edit->region.h );
}
/*
====================================================================
Handle event. Return True if message is supposed to be send (
enter pressed).
====================================================================
*/
int gui_edit_handle( GUI_Edit *edit, SDL_Event *event, int *id )
{
    int redraw = 0, ret = 0;
    if ( event->type == SDL_KEYDOWN ) {
        if ( event->key.keysym.sym == SDLK_RETURN )
            ret = 1;
        else
            if ( event->key.keysym.sym == SDLK_BACKSPACE ) {
                if ( edit->length > 0 ) {
                    edit->text[--edit->length] = 0;
                    redraw = 1;
                }
            }
            else
                if ( event->key.keysym.unicode >= 32 && event->key.keysym.unicode <= 128 )
                    if ( edit->length < EDIT_LIMIT ) {
                        edit->text[edit->length++] = event->key.keysym.unicode;
                        edit->text[edit->length] = 0;
                        redraw = 1;
                    }
    }
    if ( redraw )
        gui_edit_draw( edit );
    if ( ret ) *id = edit->id;
    return ret;
}

/*
====================================================================
GUI
====================================================================
*/
int gui_create()
{
    int sx, sy;
    char *str_diff[] = { "Easy", "Medium", "Hard" };
    if ( ( gui = calloc( 1, sizeof( GUI ) ) ) == 0 )
        goto failure;
    /* windows */
    if ( ( gui->chat_win = gui_list_create( ID_CHAT_WINDOW, 10, 4, 452, 300, 2, 13, 0 ) ) == 0 )
        goto failure;
    if ( ( gui->chat_edit = gui_edit_create( ID_CHAT_EDIT, gui->chat_win->region.x + 136, 
                                             gui->chat_win->region.y + gui->chat_win->region.h + 10, 316, 43, 2, 37 ) ) == 0 )
        goto failure;
    if ( ( gui->pause_win = gui_list_create( ID_PAUSE_WINDOW, 60, 60, 520, 300, 2, 13, 0 ) ) == 0 )
        goto failure;
    if ( ( gui->pause_edit = gui_edit_create( ID_PAUSE_EDIT, 146, 380, 348, 43, 2, 41 ) ) == 0 )
        goto failure;
    if ( ( gui->games = gui_list_create( ID_GAMES, gui->chat_win->region.x + 136, 
                                         gui->chat_edit->region.y + gui->chat_edit->region.h + 10, 316, 108, 2, 13, 0 ) ) == 0 )
        goto failure;
    if ( ( gui->channels = gui_list_create( ID_CHANNELS, 10, 
                                            gui->chat_win->region.y + gui->chat_win->region.h + 10,
                                            126, 160, 2, 13, 1 ) ) == 0 )
        goto failure;
    if ( ( gui->users = gui_list_create( ID_USERS, sdl.screen->w - 154 - 10, 4, 154, 198, 2, 13, 1 ) ) == 0 )
        goto failure;
    if ( ( gui->levels = gui_list_create( ID_LEVELS, sdl.screen->w - 154 - 10, 
                                          gui->users->region.y + gui->users->region.h + 10, 154, 108, 2, 13, 1 ) ) == 0 )
        goto failure;
    if ( ( gui->confirm = group_create( gui_create_frame( 220, 148 ), 160, 
                                        load_surf( "confirm_buttons.png", SDL_SWSURFACE ), 20, 20, 2, ID_OK, 
                                        sdl.screen, ( sdl.screen->w - 160 ) / 2, ( sdl.screen->h - 100 ) / 2 ) ) == 0 )
        goto failure;
    sx = gui->confirm->frame->img->img->w - 60; sy = gui->confirm->frame->img->img->h - 30;
    group_add_button( gui->confirm, ID_OK, sx, sy, 0, "Accept" ); sx += 30;
    group_add_button( gui->confirm, ID_CANCEL, sx, sy, 0, "Cancel" );
    group_hide( gui->confirm, 1 );
    /* fonts */
    gui->font = load_fixed_font( "f_small_yellow.png", 32, 96, 8, SDL_SWSURFACE );
    gui->hfont = load_fixed_font( "f_small_white.png", 32, 96, 8, SDL_SWSURFACE );
    /* game menu */
    if ( ( gui->mp_menu = menu_create( 0, MENU_LAYOUT_CENTERED, gui->levels->region.x, 
                                       gui->levels->region.y + gui->levels->region.h + 10, 154, 140, 2, 0 ) ) == 0 )
        goto failure;
    menu_add( gui->mp_menu, item_create_switch_x( "Difficulty:", HINT_DIFF, &config.mp_diff, str_diff, 3 ) );
    menu_add( gui->mp_menu, item_create_range( "Rounds:", HINT_ROUNDS, &config.mp_rounds, 1, 5, 1 ) );
    menu_add( gui->mp_menu, item_create_range( "Frags:", HINT_FRAGS, &config.mp_frags, 5, 30, 5 ) );
    menu_add( gui->mp_menu, item_create_range( "Balls:", HINT_BALLS, &config.mp_balls, 1, 10, 1 ) );
    menu_add( gui->mp_menu, item_create_edit( "Game Port:", HINT_GAME_PORT, config.client_game_port, 8 ) );
    menu_add( gui->mp_menu, item_create_separator( "" ) );
    menu_add( gui->mp_menu, item_create_action( "Challenge", HINT_CHALLENGE, ID_CHALLENGE ) ); 
    menu_add( gui->mp_menu, item_create_separator( "" ) );
    menu_add( gui->mp_menu, item_create_action( "Quit", HINT_, ID_QUIT ) ); 
    menu_set_fonts( gui->mp_menu, gui->font, gui->font, gui->hfont );
    menu_adjust( gui->mp_menu );
    return 1;
failure:
    gui_delete();
    return 0;
}
void gui_delete()
{
    if ( gui ) {
        gui_list_delete( gui->levels );
        gui_list_delete( gui->users );
        gui_list_delete( gui->chat_win );
        gui_edit_delete( gui->chat_edit );
        gui_list_delete( gui->pause_win );
        gui_edit_delete( gui->pause_edit );
        gui_list_delete( gui->games );
        gui_list_delete( gui->channels );
        free_font( &gui->font );
        free_font( &gui->hfont );
        if ( gui->mp_menu) menu_delete( gui->mp_menu );
        group_delete( &gui->confirm );
        free( gui );
    }
}
/*
====================================================================
Draw complete gui.
====================================================================
*/
void gui_draw()
{
    FULL_DEST( sdl.screen );
    FULL_SOURCE( cr_bkgnd );
    blit_surf();
    gui_list_draw( gui->levels );
    gui_list_draw( gui->users );
    gui_list_draw( gui->chat_win );
    gui_edit_draw( gui->chat_edit );
    gui_list_draw( gui->games );
    gui_list_draw( gui->channels );
}
/*
====================================================================
Handle an SDL event and return True if something was clicked.
====================================================================
*/
int gui_handle_event( SDL_Event *event, int *id )
{
    int ret = 0;
    if ( event->type == SDL_MOUSEBUTTONUP && last_button ) {
        if ( !last_button->lock ) {
            last_button->down = 0;
            if ( last_button->active )
                last_button->button_rect.x = 0;
        }
        if ( last_button->button_rect.x == 0 )
            if ( button_focus( last_button, event->button.x, event->button.y ) )
                last_button->button_rect.x = last_button->surf_rect.w;
        last_button = 0;
    }
    if ( motion )
        group_handle_motion( gui->confirm, motion_x, motion_y );
    if ( gui_list_handle( gui->levels, event, id ) ) ret = 1;
    else
    if ( gui_list_handle( gui->users, event, id ) ) ret = 1;
    else
    if ( gui_list_handle( gui->chat_win, event, id ) ) ret = 1;
    else
    if ( gui_edit_handle( gui->chat_edit, event, id ) ) ret = 1;
    else
    if ( gui_list_handle( gui->games, event, id ) ) ret = 1;
    else
    if ( gui_list_handle( gui->channels, event, id ) ) ret = 1;
    else
    if ( ( *id = menu_handle_event( gui->mp_menu, event ) ) != ACTION_NONE ) ret = 1;
    else
    if ( event->type == SDL_MOUSEBUTTONDOWN )
        if ( group_handle_button( gui->confirm, event->button.button, event->button.x, event->button.y, &last_button ) ) {
            ret = 1;
            *id = last_button->id;
        }
    return ret;
}
/*
====================================================================
Update levelset list
====================================================================
*/
void gui_update_levelsets()
{
    char *name;
    List *list = list_create( LIST_AUTO_DELETE, 0 );
    list_reset( levelset_names );
    while ( ( name = list_next( levelset_names ) ) )
        if ( levelset_is_network( name ) )
            list_add( list, strdup( name ) );
    gui_list_set_items( gui->levels, list );
}
/*
====================================================================
Add message to chatwindow and redraw it.
====================================================================
*/
void gui_add_chat_msg( GUI_List *chat_win, int width, char *msg, int server_info )
{
    char str[width + 2];
    int i;
    Text *text = create_text( msg, width );
    List *list = list_create( LIST_AUTO_DELETE, 0 );
    for ( i = 0; i < text->count; i++ ) {
        if ( server_info ) {
            strcpy( str + 1, text->lines[i] );
            str[0] = '!';
            list_add( list, strdup( str ) );
        }
        else
            list_add( list, strdup( text->lines[i] ) );
    }
    gui_list_add_items( chat_win, list, 200 );
    delete_text( text );
    /* adjust position so new message is seen */
    chat_win->offset = chat_win->entries->count - chat_win->length;
    if ( chat_win->offset < 0 ) chat_win->offset = 0;
    gui_list_draw( chat_win );
}
