/***************************************************************************
                          gui.h  -  description
                             -------------------
    begin                : Sun Apr 28 12:02:57 CEST 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/*
====================================================================
This is the gui for the chatroom.
====================================================================
*/

/*
====================================================================
Button
====================================================================
*/
typedef struct {
    SDL_Rect    region;
    int         id;
    SDL_Surface *image;
    int         state;
} GUI_Button;
/*
====================================================================
Create/delete button. The passed image is just  pointer.
====================================================================
*/
GUI_Button* gui_button_create( int id, int x, int y, int w, int h, SDL_Surface *surf );
void gui_button_delete( GUI_Button *button );
/*
====================================================================
Draw and handle button. Handling returns True if button was clicked.
====================================================================
*/
void gui_button_draw( GUI_Button *button );
int gui_button_handle( GUI_Button *button, SDL_Event *event, int *id );

/*
====================================================================
List (single-select)
====================================================================
*/
typedef struct {
    SDL_Rect    region;
    int         id;
    int         border;
    List        *entries;
    char        *cur_item;
    int         offset; /* first entry displayed */
    int         length; /* number of entries displayed */
    int         cell_h; /* height of an item (includes gap) */
    int         scroll_w;
    int         scroll_h; /* size of up/down/track buttons */
    int         track_h; /* track bar height */
    int         track_y; /* position in track bar */
    int         track_offset; /* absolute y where track starts */
    GUI_Button  *up, *down, *track;
    Font        *font, *hfont;
    int         select; /* if true items may be selected */
} GUI_List;
/*
====================================================================
Create/delete a list and use default buttons and fonts.
If 'select' is false this is just a memo box.
====================================================================
*/
GUI_List *gui_list_create( int id, int x, int y, int w, int h, int border, int cell_h, int select );
void gui_list_delete( GUI_List *list );
/*
====================================================================
Set list items. The old list is deleted and the new one is
used instead.
====================================================================
*/
void gui_list_set_items( GUI_List *list, List *items );
/*
====================================================================
Add items to list and remove old items if they would exceed limit
(for chatwindow). The passed list is deleted when finished.
====================================================================
*/
void gui_list_add_items( GUI_List *list, List *items, int limit );
/*
====================================================================
Draw listbox
====================================================================
*/
void gui_list_draw( GUI_List *list );
/*
====================================================================
Handle event and update the list in screen and add refresh rects.
Return True if selection has changed.
====================================================================
*/
int gui_list_handle( GUI_List *list, SDL_Event *event, int *id );

/*
====================================================================
Chat edit
====================================================================
*/
enum { EDIT_LIMIT = 111 };
typedef struct {
    SDL_Rect region;
    int id;
    int border;
    int line_w; /* characters per line */
    char text[EDIT_LIMIT + 1];
    int length;
    Font *font;
} GUI_Edit;
/*
====================================================================
Create/delete edit
====================================================================
*/
GUI_Edit *gui_edit_create( int id, int x, int y, int w, int h, int border, int lw );
void gui_edit_delete( GUI_Edit *edit );
/*
====================================================================
Draw edit
====================================================================
*/
void gui_edit_draw( GUI_Edit *edit );
/*
====================================================================
Handle event. Return True if message is supposed to be send (
enter pressed).
====================================================================
*/
int gui_edit_handle( GUI_Edit *edit, SDL_Event *event, int *id );

/*
====================================================================
Chatroom GUI
====================================================================
*/
enum { CHAT_WIN_LINE_WIDTH = 50 };
enum {
    ID_NONE = 0,
    ID_QUIT,
    ID_CHALLENGE,
    ID_USERS,
    ID_LEVELS,
    ID_GAMES,
    ID_CHANNELS,
    ID_CHAT_WINDOW,
    ID_CHAT_EDIT,
    ID_PAUSE_WINDOW,
    ID_PAUSE_EDIT,
    ID_EDIT,
    ID_OK,
    ID_CANCEL
};
typedef struct {
    GUI_Button *quit, *challenge;
    GUI_List   *users, *levels, *games, *chat_win, *channels;
    GUI_Edit   *chat_edit;
    int         state;
    Font       *font, *hfont;
    Menu       *mp_menu;
    Group      *confirm; /* window to handle the challenge */
    /* pause chatroom */
    GUI_List   *pause_win;
    GUI_Edit   *pause_edit;
} GUI;
int gui_create();
void gui_delete();
/*
====================================================================
Draw complete gui.
====================================================================
*/
void gui_draw();
/*
====================================================================
Handle an SDL event and return True if something was clicked.
====================================================================
*/
int gui_handle_event( SDL_Event *event, int *id );
/*
====================================================================
Update levelset list
====================================================================
*/
void gui_update_levelsets();
/*
====================================================================
Add message to chatwindow and redraw it.
====================================================================
*/
void gui_add_chat_msg( GUI_List *chat_win, int width, char *msg, int server_info );
