/***************************************************************************
                          shots.c  -  description
                             -------------------
    begin                : Sat Sep 8 2001
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lbreakout.h"
#include "config.h"
#include "levels.h"
#include "shrapnells.h"
#include "player.h"
#include "display.h"
#include "paddle.h"
#include "bricks.h"
#include "shots.h"
#ifdef SOUND
#include "audio.h"
#endif

extern SDL_Surface *shot_pic;
extern SDL_Surface *shot_shadow;
List *shots; /* list of all shots */
int shot_w = 10;
int shot_h = 10;
int shot_fr_num = 4;
float shot_fpms = 0.01;
float shot_v_y = 0.2;
int shot_alpha = 64;
extern Brick bricks[MAP_WIDTH][MAP_HEIGHT];
extern SDL_Surface *offscreen;
extern Sdl sdl;
extern int shadow_size;
extern Config config;
#ifdef SOUND
extern Sound_Chunk *wav_shot;
#endif
extern int game_type, game_host;
extern Paddle *paddles[4];
extern int paddle_count;
extern int net_shot_fired;

/*
====================================================================
Locals
====================================================================
*/

/*
====================================================================
Compute target of shot.
====================================================================
*/
void shot_get_target( Shot *shot ) {
    int mx = (int)(shot->x + 3) / BRICK_WIDTH;
    int my = (int)(shot->y + 3 + ((shot->dir==1)?5:0) ) / BRICK_HEIGHT;
    memset( &shot->t, 0, sizeof(Target) );
    while ( bricks[mx][my].id == -1 ) 
        my += shot->dir;
    shot->t.mx = mx; shot->t.my = my;
    mx = (int)(shot->x + 6) / BRICK_WIDTH;
    if (mx != shot->t.mx) {
        my = (int)(shot->y + 3 + ((shot->dir==1)?5:0) ) / BRICK_HEIGHT;
        while(bricks[mx][my].id == -1) 
            my += shot->dir;
        if (my == shot->t.my)
            shot->next_too = 1;
        else
            if ( ( shot->dir == -1 && my > shot->t.my ) ||
                 ( shot->dir ==  1 && my < shot->t.my ) ) {
                shot->t.mx = mx;
                shot->t.my = my;
                shot->next_too = 0;
            }
    }
    shot->t.cur_tm = 0;
    shot->t.time = abs( (int)((shot->y + 3 + ((shot->dir==1)?5:0) - (shot->t.my * BRICK_HEIGHT + ((shot->dir==-1)?(BRICK_HEIGHT - 1):0) )) / fabs(shot_v_y)) );
}

/*
====================================================================
Publics
====================================================================
*/

/*
====================================================================
Load, delete shot graphics
====================================================================
*/
void shot_load()
{
    shots = list_create( LIST_AUTO_DELETE, LIST_NO_CALLBACK );
}
void shot_delete()
{
    if ( shots ) list_delete( shots ); shots = 0;
}
/*
====================================================================
Create new shot at position (centered).
'signum' of direction determines into which direction the shot 
vertically goes.
====================================================================
*/
void shot_create( Paddle *paddle )
{
    Shot *shot = calloc( 1, sizeof( Shot ) );
    shot->cur_fr = 0;
    shot->paddle = paddle;
    shot->dir = (paddle->type == PADDLE_TOP) ? 1 : -1;
    shot->x = paddle->x + ( paddle->w >> 1 ) - (shot_w >> 1);
    shot->y = paddle->y + ( paddle->h >> 1 ) - (shot_h >> 1);
    shot_get_target( shot );
    list_add( shots, shot );
#ifdef SOUND
    sound_play( wav_shot );
    net_shot_fired = 1;
#endif
}
/*
====================================================================
Delete all shots
====================================================================
*/
void shots_reset()
{
    list_clear( shots );
}
/*
====================================================================
Hide and show shots
====================================================================
*/
void shots_hide()
{
    List_Entry  *entry = shots->head.next;
    Shot        *shot;
    int w = shot_w, h = shot_h;
    if ( config.shadow ) { w += shadow_size; h += shadow_size; }
    while ( entry != &shots->tail ) {
        shot = entry->item;
        DEST(sdl.screen, (int)shot->x, (int)shot->y, w, h );
        SOURCE(offscreen, (int)shot->x, (int)shot->y);
        blit_surf();
        add_refresh_rect((int)shot->x, (int)shot->y, w, h );
        entry = entry->next;
    }
}
void shots_show()
{
    List_Entry  *entry = shots->head.next;
    Shot        *shot;
    if ( config.shadow ) set_surf_clip( sdl.screen, 0, 0, sdl.screen->w - BRICK_WIDTH, sdl.screen->h );
    while ( entry != &shots->tail ) {
        shot = entry->item;
        if ( config.shadow ) {
            DEST(sdl.screen, (int)shot->x + shadow_size, (int)shot->y + shadow_size, shot_w, shot_h);
            SOURCE(shot_shadow, 0, 0);
            alpha_blit_surf(SHADOW_ALPHA);
        }
        DEST(sdl.screen, (int)shot->x, (int)shot->y, shot_w, shot_h);
        SOURCE(shot_pic, (int)shot->cur_fr * shot_w, 0);
        alpha_blit_surf(shot_alpha);
        if ( config.shadow )
            add_refresh_rect((int)shot->x, (int)shot->y, shot_w + shadow_size, shot_h + shadow_size);
        else
            add_refresh_rect((int)shot->x, (int)shot->y, shot_w, shot_h);
        entry = entry->next;
    }
    if ( config.shadow ) set_surf_clip( sdl.screen, 0, 0, 0, 0 );
}
void shots_alphashow( int alpha )
{
    List_Entry  *entry = shots->head.next;
    Shot        *shot;
    while ( entry != &shots->tail ) {
        shot = entry->item;
        DEST(sdl.screen, (int)shot->x, (int)shot->y, shot_w, shot_h);
        SOURCE(shot_pic, (int)shot->cur_fr * shot_w, 0);
        alpha_blit_surf(alpha);
        add_refresh_rect((int)shot->x, (int)shot->y, shot_w, shot_h);
        entry = entry->next;
    }
}
/*
====================================================================
Update position of shots and check if bricks get destroyed.
A list of all destroyed bricks is returned (at maximum 
PADDLE_WEAPON_AMMO * 4)
====================================================================
*/
void shots_update( int ms, Brick_Pos *bricks, int *count )
{
    int i;
    List_Entry  *entry = shots->head.next;
    Shot        *shot;
    *count = 0;
    while ( entry != &shots->tail ) {
        shot = entry->item;
        shot->cur_fr += ms * shot_fpms;
        if (shot->cur_fr >= shot_fr_num) shot->cur_fr -= shot_fr_num;
        shot->y += shot->dir * ms * shot_v_y;
        shot->t.cur_tm += ms;
        entry = entry->next;
        /* kill 'out of screen' shots */
        if ( shot->y + shot_h < 0 || shot->y > sdl.screen->h ) {
            shot->paddle->weapon_ammo++; /* give back used shot */
            list_delete_entry( shots, entry->prev );
            continue;
        }
        /* check hits */
        if (shot->t.cur_tm > shot->t.time) {
            if ( brick_hit(shot->t.mx, shot->t.my, 0, SHR_BY_SHOT, vector_get( 0, shot->dir ), shot->paddle ) ) {
                shots_get_new_targets( shot->t.mx, shot->t.my );
                bricks[*count].x = shot->t.mx;
                bricks[(*count)++].y = shot->t.my;
            }
            if (shot->next_too) {
                if ( brick_hit(shot->t.mx + 1, shot->t.my, 0, SHR_BY_SHOT, vector_get( 0, shot->dir ), shot->paddle ) ) {
                    shots_get_new_targets(shot->t.mx + 1, shot->t.my);
                    bricks[*count].x = shot->t.mx + 1;
                    bricks[(*count)++].y = shot->t.my;
                }
            }
            shot->paddle->weapon_ammo++; /* give back used shot */
            list_delete_entry( shots, entry->prev );
            continue;
        }
        /* in multiplayer we check if we hit the opponent if so we steal
           him a 1000 points */
        if ( game_type == GT_NETWORK )
            for ( i = 0; i < paddle_count; i++ )
                if ( paddles[i] != shot->paddle )
                if ( shot->x + shot_w > paddles[i]->x )
                if ( shot->x < paddles[i]->x + paddles[i]->w )
                if ( shot->y + shot_h > paddles[i]->y )
                if ( shot->y < paddles[i]->y + paddles[i]->h ) {
                    player_add_score( paddles[i]->player, -1000 );
                    player_add_score( shot->paddle->player, 1000 );
                    shot->paddle->weapon_ammo++;
                    list_delete_entry( shots, entry->prev );
                }
    }
}
/*
====================================================================
Get new targets for all shots that where supposed to
destroy brick at x,y
====================================================================
*/
void shots_get_new_targets( int x, int y )
{
    Shot        *shot;
    list_reset( shots );
    while ( ( shot = list_next( shots ) ) )
        if (shot->t.mx == x && shot->t.my == y)
            shot_get_target( shot );
}
