/***************************************************************************
                          windows.c  -  description
                             -------------------
    begin                : Tue Mar 21 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "lbreakout.h"
#include "file.h"
#include "event.h"
#include "windows.h"

enum {
    ID_INTERN_UP = -101,
    ID_INTERN_DOWN = -100
};

/*
====================================================================
Externals
====================================================================
*/
extern Sdl sdl;

/*
====================================================================
Create button group. At maximum 'limit' buttons may be added.
A buttons tooltip is displayed in 'label'.
An actual button id is computed as base_id + id.
====================================================================
*/
Group *group_create( SDL_Surface *frame, int alpha, SDL_Surface *img, int w, int h, int limit, int base_id, 
                     SDL_Surface *surf, int x, int y )
{
    Group *group = calloc( 1, sizeof( Group ) );
    if ( ( group->frame = iframe_create( frame, alpha, surf, x, y ) ) == 0 ) goto failure;
    if ( img == 0 ) {
        fprintf( stderr, "group_create: passed button surface is NULL: %s\n", SDL_GetError() );
        goto failure;
    }
    SDL_SetColorKey( img, SDL_SRCCOLORKEY, 0x0 );
    group->img = img;
    group->w = w; group->h = h;
    group->base_id = base_id;
    group->button_limit = limit;
    if ( ( group->buttons = calloc( group->button_limit, sizeof( Button ) ) ) == 0 ) {
        fprintf( stderr, "Out of memory\n" );
        goto failure;
    }
    return group;
failure:
    group_delete( &group );
    return 0;
}
void group_delete( Group **group )
{
    if ( *group ) {
        iframe_delete( &(*group)->frame );
        if ( (*group)->img ) SDL_FreeSurface( (*group)->img );
        if ( (*group)->buttons ) free( (*group)->buttons );
        free( *group ); *group = 0;
    }
}

/*
====================================================================
Add a button at x,y in group. If lock is true this button is a switch.
'id' * group::h is the y_offset of the button.
====================================================================
*/
int group_add_button( Group *group, int id, int x, int y, int lock, char *tooltip )
{
    if ( group->button_count == group->button_limit ) {
        fprintf( stderr, "This group has reached it'S maximum number of buttons.\n" );
        return 0;
    }
    group->buttons[group->button_count].surf_rect.x = x + group->frame->img->bkgnd->surf_rect.x;
    group->buttons[group->button_count].surf_rect.y = y + group->frame->img->bkgnd->surf_rect.y;
    group->buttons[group->button_count].surf_rect.w = group->w;
    group->buttons[group->button_count].surf_rect.h = group->h;
    group->buttons[group->button_count].button_rect.x = 0;
    group->buttons[group->button_count].button_rect.y = ( id - group->base_id ) * group->h;
    group->buttons[group->button_count].button_rect.w = group->w;
    group->buttons[group->button_count].button_rect.h = group->h;
    group->buttons[group->button_count].id = id;
    group->buttons[group->button_count].active = 1;
    group->buttons[group->button_count].lock = lock;
    group->button_count++;
    return 1;
}

/*
====================================================================
Add a button at x,y in group. If lock is true this button is a switch.
'icon_id' is used for the button icon instead of 'id' (allows
multiple buttons of the same icon)
====================================================================
*/
int group_add_button_complex( Group *group, int id, int icon_id, int x, int y, int lock, char *tooltip )
{
    if ( group->button_count == group->button_limit ) {
        fprintf( stderr, "This group has reached it'S maximum number of buttons.\n" );
        return 0;
    }
    group->buttons[group->button_count].surf_rect.x = x + group->frame->img->bkgnd->surf_rect.x;
    group->buttons[group->button_count].surf_rect.y = y + group->frame->img->bkgnd->surf_rect.y;
    group->buttons[group->button_count].surf_rect.w = group->w;
    group->buttons[group->button_count].surf_rect.h = group->h;
    group->buttons[group->button_count].button_rect.x = 0;
    group->buttons[group->button_count].button_rect.y = icon_id * group->h;
    group->buttons[group->button_count].button_rect.w = group->w;
    group->buttons[group->button_count].button_rect.h = group->h;
    group->buttons[group->button_count].id = id;
    group->buttons[group->button_count].active = 1;
    group->buttons[group->button_count].lock = lock;
    group->button_count++;
    return 1;
}

/*
====================================================================
Get button by global id.
====================================================================
*/
Button* group_get_button( Group *group, int id )
{
    int i;
    for ( i = 0; i < group->button_count; i++ )
        if ( group->buttons[i].id == id )
            return &group->buttons[i];
    return 0;
}
/*
====================================================================
Set active state of a button by global id.
====================================================================
*/
void group_set_active( Group *group, int id, int active )
{
    int i;
    for ( i = 0; i < group->button_count; i++ )
        if ( id == group->buttons[i].id ) {
            group->buttons[i].active = active;
            if ( active ) {
               group->buttons[i].button_rect.x = 0;
               group->buttons[i].down = 0;
            }
            else
               group->buttons[i].button_rect.x = 3 * group->w;
            break;
        }
}

/*
====================================================================
Lock button.
====================================================================
*/
void group_lock_button( Group *group, int id, int down )
{
    Button *button = group_get_button( group, id );
    if ( button && button->lock ) {
        button->down = down;
        if ( down )
            button->button_rect.x = 2 * group->w;
        else
            button->button_rect.x = 0;
    }
}
/*
====================================================================
Check motion event and modify buttons and label.
Return True if in this frame.
====================================================================
*/
int group_handle_motion( Group *group, int x, int y )
{
    int i;
    if ( !group->frame->img->bkgnd->hide )
    if ( x >= group->frame->img->bkgnd->surf_rect.x && y >= group->frame->img->bkgnd->surf_rect.y )
    if ( x < group->frame->img->bkgnd->surf_rect.x + group->frame->img->bkgnd->surf_rect.w )
    if ( y < group->frame->img->bkgnd->surf_rect.y + group->frame->img->bkgnd->surf_rect.h ) {
        for ( i = 0; i < group->button_count; i++ )
            if ( group->buttons[i].active ) {
                if ( button_focus( &group->buttons[i], x, y ) ) {
                    if ( !group->buttons[i].down )
                        group->buttons[i].button_rect.x = group->w;
                }
                else
                    if ( !group->buttons[i].down || !group->buttons[i].lock )
                        group->buttons[i].button_rect.x = 0;
            }
        return 1;
    }
    for ( i = 0; i < group->button_count; i++ )
        if ( group->buttons[i].active && ( !group->buttons[i].down || !group->buttons[i].lock ) )
            group->buttons[i].button_rect.x = 0;
    return 0;
}
/*
====================================================================
Check click event. Return True if button was clicked and return
the button id.
====================================================================
*/
int group_handle_button( Group *group, int button_id, int x, int y, Button **button )
{
    int i;
    if ( button_id == LEFT_BUTTON ) 
    if ( !group->frame->img->bkgnd->hide )
    if ( x >= group->frame->img->bkgnd->surf_rect.x && y >= group->frame->img->bkgnd->surf_rect.y )
    if ( x < group->frame->img->bkgnd->surf_rect.x + group->frame->img->bkgnd->surf_rect.w )
    if ( y < group->frame->img->bkgnd->surf_rect.y + group->frame->img->bkgnd->surf_rect.h ) {
        for ( i = 0; i < group->button_count; i++ )
            if ( group->buttons[i].active )
                if ( button_focus( &group->buttons[i], x, y ) ) {
                    *button = &group->buttons[i];
                    (*button)->down = !(*button)->down;
                    if ( (*button)->down )
                        (*button)->button_rect.x = 2 * group->w;
                    else
                        (*button)->button_rect.x = 0;
                    return 1;
                }
    }
    return 0;
}

/*
====================================================================
Draw group.
====================================================================
*/
void group_draw( Group *group )
{
    int i;
    iframe_draw( group->frame );
    if ( !group->frame->img->bkgnd->hide )
        for ( i = 0; i < group->button_count; i++ )
            SDL_BlitSurface( group->img, &group->buttons[i].button_rect,
                             group->frame->img->bkgnd->surf, &group->buttons[i].surf_rect );
}

/*
====================================================================
Modify settings.
====================================================================
*/
void group_move( Group *group, int x, int y )
{
    int i;
    int rel_x = x - group->frame->img->bkgnd->surf_rect.x;
    int rel_y = y - group->frame->img->bkgnd->surf_rect.y;
    iframe_move( group->frame, x, y );
    for ( i = 0; i < group->button_count; i++ ) {
        group->buttons[i].surf_rect.x += rel_x;
        group->buttons[i].surf_rect.y += rel_y;
    }
}

/*
====================================================================
Return True if x,y is on group button.
====================================================================
*/
int button_focus( Button *button, int x, int y )
{
    if ( x >= button->surf_rect.x && y >= button->surf_rect.y )
        if ( x < button->surf_rect.x + button->surf_rect.w && y < button->surf_rect.y + button->surf_rect.h )
            return 1;
    return 0;
}
