/***************************************************************************
                          windows.h  -  description
                             -------------------
    begin                : Tue Mar 21 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __WINDOWS_H
#define __WINDOWS_H

#include "image.h"

/*
====================================================================
Button group. A frame with some click buttons.
====================================================================
*/
typedef struct {
    SDL_Rect button_rect;   /* region in button surface */
    SDL_Rect surf_rect;     /* region in buffer::surface */ 
    int id;                 /* if returned on click */
    char tooltip[32];       /* displayed as short help */
    int active;             /* true if button may be used */
    int down;               /* true if currently pressed */
    int lock;               /* keep button down when released */
} Button;
typedef struct {
    Frame *frame;       /* frame of group */
    SDL_Surface *img;   /* button surface. a row is a button, a column is a state */
    int w, h;           /* button size */
    Button *buttons;    
    int button_count;   /* number of buttons */
    int button_limit;   /* limit of buttons */
    int base_id; 
} Group;

/*
====================================================================
Create button group. At maximum 'limit' buttons may be added.
A buttons tooltip is displayed in 'label'.
An actual button id is computed as base_id + id.
====================================================================
*/
Group *group_create( SDL_Surface *frame, int alpha, SDL_Surface *img, int w, int h, int limit, int base_id, 
                     SDL_Surface *surf, int x, int y );
void group_delete( Group **group );

/*
====================================================================
Add a button at x,y in group. If lock is true this button is a switch.
'id' * group::h is the y_offset of the button.
====================================================================
*/
int group_add_button( Group *group, int id, int x, int y, int lock, char *tooltip );

/*
====================================================================
Add a button at x,y in group. If lock is true this button is a switch.
'icon_id' is used for the button icon instead of 'id' (allows
multiple buttons of the same icon)
====================================================================
*/
int group_add_button_complex( Group *group, int id, int icon_id, int x, int y, int lock, char *tooltip );

/*
====================================================================
Get button by global id.
====================================================================
*/
Button* group_get_button( Group *group, int id );

/*
====================================================================
Set active state of a button by global id.
====================================================================
*/
void group_set_active( Group *group, int id, int active );

/*
====================================================================
Lock button.
====================================================================
*/
void group_lock_button( Group *group, int id, int down );

/*
====================================================================
Check motion event and modify buttons and label.
Return True if in this frame.
====================================================================
*/
int group_handle_motion( Group *group, int x, int y );
/*
====================================================================
Check click event. Return True if button was clicked and return
the button.
====================================================================
*/
int group_handle_button( Group *group, int button_id, int x, int y, Button **button );

/*
====================================================================
Draw group.
====================================================================
*/
#define group_hide( group, hide ) buffer_hide( group->frame->img->bkgnd, hide )
#define group_get_bkgnd( group ) buffer_get( group->frame->img->bkgnd )
#define group_draw_bkgnd( group ) buffer_draw( group->frame->img->bkgnd )
void group_draw( Group *group );

/*
====================================================================
Modify settings.
====================================================================
*/
#define group_set_surface( group, surf ) buffer_set_surface( group->frame->img->bkgnd, surf )
void group_move( Group *group, int x, int y );

/*
====================================================================
Return True if x,y is on group button.
====================================================================
*/
int button_focus( Button *button, int x, int y );

#endif
