/***************************************************************************
                          server.h  -  description
                             -------------------
    begin                : Sun Apr 28 12:02:57 CEST 2002
    copyright            : (C) 2001 by Michael Speck
    email                : kulkanie@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef WITH_NETWORK

/*
====================================================================
Definitions used by server and client.
====================================================================
*/

/*
====================================================================
Protocol used by server and client (must be equal)
====================================================================
*/
#define PROTOCOL 1

/*
====================================================================
Server/client messages.
====================================================================
*/
enum {
    MSG_NONE = 0,
    MSG_ERROR,          /* transfer an error message */
    MSG_SERVER_INFO,    /* server message written to chat */
    MSG_IDENTIFY,       /* request nickname and lbreakout2 version */
    MSG_LOGIN_OKAY,     /* accept login */
    MSG_CHAT,           /* a chat message */
    MSG_PREPARE_UPDATE, /* tell client to prepare for full update */
    MSG_ADD_USER,       /* add a user to client's challengable list */
    MSG_REMOVE_USER,    /* remove a user from client's challengable list */
    MSG_ADD_USERS,      /* packed channels to speed up full updates */
    MSG_ADD_GAME,       /* add a game analoguely */
    MSG_REMOVE_GAME,    /* dito */
    MSG_UPDATE_GAME,    /* send update of a specific game */
    MSG_ADD_GAMES,      /* packed channels to speed up full updates */
    MSG_ADD_CHANNEL,
    MSG_REMOVE_CHANNEL, /* chat channels */
    MSG_ADD_CHANNELS,   /* packed channels to speed up full updates */
    MSG_CHALLENGE,      /* challenge another user */
    MSG_ACCEPT_CHALLENGE, /* accept challenge */
    MSG_REJECT_CHALLENGE, /* reject challenge */
    MSG_BUSY,           /* user can't accept challenges at the moment */
    MSG_CANCEL_CHALLENGE, /* cancel your challenge */
    MSG_GAME_EXITED,    /* returned if user leaves a game (either normal or abnormal) */
    MSG_DISCONNECT,     /* received when player leaves server */
    MSG_IP_ADDRESS,     /* ip address of the game host */
    MSG_UNLINK,         /* cut a direct connection */
    MSG_LEVEL_COUNT,    /* number of levels in the levelset */
    MSG_LEVEL_DATA,     /* sent raw leveldata of current level to client */
    MSG_READY,          /* send if player is ready */
    MSG_PAUSE,          /* pause game */
    MSG_UNPAUSE,        /* guess what, eh? */
    MSG_BENCHMARK,      /* result of speed test (faster machine becomes host) */
    MSG_DM_HOSTDATA,    /* ingame package send by host (deathmatch) */
    MSG_DM_CLIENTDATA,  /* ingame message send by client */
    MSG_MP_HOSTDATA,    /* ingame package send by host (multiplayer) */
    MSG_MP_CLIENTDATA,  /* ingame message send by client */
    MSG_ROUND_OVER,     /* result of a network round */
    MSG_WINNER,         /* game host sends winner of round to server */
    MSG_SWITCH_CHANNEL, /* switch to this channel */
    MSG_SET_CHANNEL,    /* update clients channel id */
    MSG_INIT_TRANSFER,  /* send an transfer offer to remote user */
    MSG_CANCEL_TRANSFER,/* stop the transfer */
    MSG_ACCEPT_SET,     
    MSG_REJECT_SET,     /* accept/refuse set offer */
    MSG_OLDER_SET       /* send if you want to update a newer set */
};
enum { CHAT_MSG_LIMIT = 112 };

/*
====================================================================
User
====================================================================
*/
typedef struct _SUser{
    int id;             
    char name[16];      /* nickname */
    Net_Socket *socket; /* connection endpoint to server */
    int playing;        /* true if in game */
    int ready;          /* true if user is completely accepted by server */
    int idle_start_time; /* time when last package was received */
    struct _SUser *r_user; /* if this pointer is not NULL all packages
                              are directly sent to the remote user
                              without being processed by the server */
    int channel;        /* id of the channel the user is currently in */
    int transfer_ok;    /* user will accept confirm requests */
} SUser;

/*
====================================================================
Game
====================================================================
*/
typedef struct {
    int   id;     
    int   running; /* if actually running the game will be posted  
                      (challenges create an empty SGame that's why we
                       need this flag) */
    SUser *host;  /* player who made the challenge (pointer to 'users')*/
    SUser *guest; /* player that was challenged (pointer to 'users')*/
    char  players[40]; /* a string with the player names */
    char  set[16];  /* levelset that is played */
    int   host_wins, guest_wins; /* won rounds */
} SGame;

/*
====================================================================
Channel
====================================================================
*/
typedef struct {
    int id;
    char name[16];  /* name of the channel */
    List *users;    /* list of all users in this channel */
    List *games;    /* list of all games in this channel */
} SChannel;

#endif
